import copy
import warnings
import numpy as np
import ot
from scipy import stats
from . import utilities, dist_reg, interpolation
from .utilities import BatchedCategorical
import pandas as pd
import cvxpy as cp
# typing
from scipy.stats._distn_infrastructure import rv_generic
import sklearn.base
from typing import Optional, Union

MIN_NVALS = 7
DISC_THRESH = 2 # treat vars. with <= DISC_THRESH values as discrete
CROSSFIT_WARNING = """
==================================================
Not fitting a model because y0_dists/y1_dists were
directly provided. Please ensure cross-fitting is
employed correctly, else inference will be invalid
(see https://arxiv.org/abs/2310.08115). To suppress
this warning, set ``suppress_warning=True``.
==================================================
"""

def get_default_model(discrete, support, outcome_model=None, **model_kwargs):
	# Prevent errors when eps_dist is provided for discrete data
	if discrete:
		model_kwargs.pop("eps_dist", None)

	# Handle the case where we have a list of outcome models
	if isinstance(outcome_model, list):
		return [
			get_default_model(outcome_model=x, discrete=discrete, support=support, **model_kwargs)
			for x in outcome_model
		]
	if isinstance(outcome_model, dist_reg.DistReg):
		return outcome_model
	outcome_model = 'ridge' if outcome_model is None else outcome_model
	if not discrete:
		return dist_reg.CtsDistReg(
			model_type=outcome_model, **model_kwargs
		)
	elif discrete and set(support) == set([0, 1]):
		return dist_reg.BinaryDistReg(
			model_type=outcome_model, **model_kwargs
		)
	else:
		raise NotImplementedError("Currently no default for non-binary discrete data")

def infer_discrete(discrete, support, y):
	n = len(y)
	### Check if discrete
	if n <= 10:
		if discrete is None:
			raise ValueError("Please specify the value of discrete as n <= 10")
		if discrete and support is None:
			raise ValueError("Please specify the value of support as n <= 10")
	if support is None:
		support = np.unique(y)
	if discrete is None:
		if len(support) <= DISC_THRESH:
			discrete = True
		else:
			discrete = False

	# Adjust support to avoid being misleading
	# in the continuous case
	if not discrete:
		support = None
	return discrete, support

def _default_ylims(
	y, 
	y0_min=None,
	y1_min=None,
	y0_max=None,
	y1_max=None,
	**kwargs # ignored
):
	# Computes default bounds on y
	if y0_min is None:
		y0_min = 1.5 * y.min() - y.max() / 2
	if y1_min is None:
		y1_min = 1.5 * y.min() - y.max() / 2
	if y0_max is None:
		y0_max = 1.5 * y.max() - y.min() / 2
	if y1_max is None:
		y1_max = 1.5 * y.max() - y.min() / 2
	return y0_min, y1_min, y0_max, y1_max

def _dualvals_are_too_large(nu0, nu1, almost_inf):
	return max(np.max(np.abs(nu0)), np.max(np.abs(nu1))) > 0.1 * almost_inf

class DualBounds:
	"""
	Computes dual bounds on :math:`E[f(Y(0),Y(1), X)].`

	Here, :math:`X` are covariates and :math:`Y(0), Y(1)` are 
	potential outcomes.

	Parameters
	----------
	f : function
		Function which defines the partially identified estimand.
		Must be a function of three arguments: y0, y1, x 
		(in that order). E.g.,
		``f = lambda y0, y1, x : y0 <= y1``
	outcome : np.array | pd.Series
		n-length array of outcome measurements (Y).
	treatment : np.array | pd.Series
		n-length array of binary treatment (W).
	covariates : np.array | pd.Series
		(n, p)-shaped array of covariates (X).
	propensities : np.array | pd.Series
		n-length array of propensity scores :math:`P(W=1 | X)`. 
		If ``None``, will be estimated from the data.
	clusters : np.array | pd.Series
		Optional n-length array of clusters, so ``clusters[i] = j``
		indicates that observation i is in cluster j.
	outcome_model : str | dist_reg.DistReg | list
		The model for estimating the law of :math:`Y | X, W`.
		Three options:

		- A str identifier, e.g., 'ridge', 'lasso', 'elasticnet', 'randomforest', 'knn'.
		- An object inheriting from ``dist_reg.DistReg``. 
		- A list of ``dist_reg.DistReg`` objects to automatically choose between.

		E.g., when ``outcome`` is continuous, the default is
		``outcome_model=dist_reg.CtsDistReg(model_type='ridge')``.
	propensity_model : str | sklearn classifier
		How to estimate the propensity scores if they are not provided.
		Two options:

		- A str identifier, e.g., 'ridge', 'lasso', 'elasticnet', 'randomforest', 'knn'.
		- An sklearn classifier, e.g., ``sklearn.linear_model.LogisticRegressionCV()``.
	model_selector : dist_reg.ModelSelector
		A ModelSelector object which can choose between several outcome models.
		The default performs within-fold nested cross-validation. Note: this
		argument is ignored unless ``outcome_model`` is a list.
	discrete : bool
		If True, treats the outcome as a discrete variable. 
		Defaults to ``None`` (inferred from the data).
	support : np.array
		Optional support of the outcome, if known and discrete.
		Defaults to ``None`` (inferred from the data).
	support_restriction : function
		Boolean-valued function of y0, y1, x where 
		``support_restriction(y0, y1, x) = False`` asserts that 
		y0, y1, x is not in the support of :math:`Y(0), Y(1), X`.
		Defaults to ``None`` (no a-priori support restrictions).
		See the user guide for important usage tips.
	model_kwargs : dict
		Additional kwargs for the ``outcome_model``, e.g.,
		``feature_transform``. See 
		:class:`dualbounds.dist_reg.CtsDistReg` or 
		:class:`dualbounds.dist_reg.BinaryDistReg` for more kwargs.

	Notes
	-----
	``DualBounds`` will do limited preprocessing to (e.g.) create 
	dummies for discrete covariates. However, we recommended doing
	custom preprocessing for optimal results.

	Examples
	--------
	Here we fit DualBounds on :math:`P(Y(0) < Y(1))` based on
	synthetic regression data: ::
		import dualbounds as db

		# Generate synthetic data
		data = db.gen_data.gen_regression_data(n=900, p=30)

		# Initialize dual bounds object
		dbnd = db.generic.DualBounds(
			f=lambda y0, y1, x: y0 < y1,
			covariates=data['X'],
			treatment=data['W'],
			outcome=data['y'],
			propensities=data['pis'],
			outcome_model='ridge',
		)

		# Compute dual bounds and observe output
		dbnd.fit(alpha=0.05).summary()
	"""
	def __init__(
		self, 
		f: callable,
		outcome: Union[np.array, pd.Series],
		treatment: Union[np.array, pd.Series],
		covariates: Optional[Union[np.array, pd.DataFrame]]=None,
		propensities: Optional[Union[np.array, pd.Series]]=None,
		clusters: Optional[Union[np.array, pd.Series]]=None,
		outcome_model: Union[str, dist_reg.DistReg, list]='ridge',
		propensity_model: Union[str, sklearn.base.BaseEstimator]='ridge',
		model_selector: Optional[dist_reg.ModelSelector]=None,
		discrete: Optional[np.array]=None,
		support: Optional[np.array]=None,
		support_restriction: Optional[callable]=None,
		**model_kwargs,
	) -> None:
		# Estimand
		self.f = f

		### Process outcome
		self.y = outcome
		if isinstance(self.y, pd.Series) or isinstance(self.y, pd.DataFrame):
			self.y = self.y.values
		if np.any(np.isnan(self.y)):
			raise ValueError("outcome contains nans")
		if len(self.y.shape) > 1:
			if len(self.y.shape) == 2 and self.y.shape[1] == 1:
				self.y = self.y.flatten()
			else:
				raise ValueError("outcome should be a flat array not a matrix")
		self.n = len(self.y)

		### Treatment
		if isinstance(treatment, pd.Series) or isinstance(treatment, pd.DataFrame):
			treatment = treatment.values.reshape(self.n)
		self.W = utilities._binarize_variable(treatment, var_name='treatment')

		### Process covariates
		self.X = covariates
		if self.X is None:
			self.X = np.zeros((self.n, 1))
		# limited preprocessing of covariates
		if isinstance(self.X, pd.DataFrame):
			self.X = utilities.process_covariates(self.X)
			self.cov_names = self.X.columns
			self.X = self.X.values
		else:
			if len(self.X.shape) == 1:
				self.X = self.X.reshape(-1, 1)
			self.cov_names = np.arange(self.X.shape[1])
		# fill NAs if existing
		if np.any(np.isnan(self.X)):
			self.X = self.X.copy()
			naninds = np.where(np.isnan(self.X))
			self.X[naninds] = np.take(np.nanmean(self.X, axis=0), naninds[1])

		### process propensities
		self.pis = propensities
		if isinstance(self.pis, pd.Series) or isinstance(self.pis, pd.DataFrame):
			self.pis = self.pis.values.reshape(self.n)
		if self.pis is not None:
			if np.any(np.isnan(self.pis)):
				raise ValueError(f"propensities (pis) are provided but contains missing values")
			if np.any(self.pis < 0) or np.any(self.pis > 1):
				raise ValueError(f"propensities (pis) do not lie within [0,1]")

		### Clusters
		if clusters is not None:
			if isinstance(clusters, pd.Series) or isinstance(clusters, pd.DataFrame):
				clusters = clusters.values.reshape(self.n)
		self.clusters = clusters

		### Check if discrete
		self.discrete, self.support = infer_discrete(
			discrete=discrete, support=support, y=self.y,
		)
		self.outcome_model = outcome_model
		self.propensity_model = propensity_model
		self.model_selector = model_selector
		self.model_kwargs = model_kwargs

		## Support restrictions
		self.support_restriction = support_restriction

		## Initialize core objects
		self.y0_dists = None
		self.y1_dists = None
		self.oos_dist_preds = None

	def _apply_ot_fn(
		self, fn: callable, y0: Union[float, np.array], y1: Union[float, np.array], **kwargs
	):
		"""
		Applies fn and infers appropriate broadcasting for the OT setting.

		Parameters
		----------
		y0 : float | np.array
		y1 : float | np.array
		kwargs : dict
			Other parameters, e.g., x, or w0 and w1 in the IV setting
			(since DualIVBounds inherits this function).

		Returns
		-------
		fn_val : float | np.array
			Array of shape (len(y0) x len(y1)).

		Notes
		-----
		Internally, this is used to apply self.f and self.support_restriction.
		"""
		y0_haslength = utilities.haslength(y0)
		y1_haslength = utilities.haslength(y1)
		# Handle the scalar case
		if not y0_haslength and not y1_haslength:
			return fn(y0=y0, y1=y1, **kwargs)
		# Otherwise vectorize appropriately 
		if not y0_haslength:
			y0 = np.array([y0])
		if not y1_haslength:
			y1 = np.array([y1])

		# Adding extra zeros ensures correct broadcasting
		orig = fn(y0=y0.reshape(-1, 1), y1=y1.reshape(1, -1), **kwargs)
		return orig + np.zeros((len(y0), len(y1)))
		
	def _discretize(
		self, 
		ydists,
		nvals,
		ymin,
		ymax,
		min_quantile=0.001,
		ninterp=None,
		min_yprob=1e-8,
	):
		"""
		Helper method which discretizes Y before solving +
		interpolating to obtain dual variables.
		
		Notes
		-----
		For discrete Y: extracts conditional PMF/support points.
		For continuous Y: discretizes ydists along evenly spaced
		quantiles. Also adds extra points near the edge of the
		support to ensure numerical stability.
		"""
		# allow for batched setting
		if not isinstance(ydists, list):
			ydists = [ydists]

		### Discrete case ###
		if self.discrete:
			yvals = []
			yprobs = []
			for ydist in ydists:
				if not isinstance(ydist, utilities.BatchedCategorical):
					msg = "self.discrete=True, so the predicted ydists must be a BatchedCategorical"
					msg += "distributions. To get discrete-like behavior using other distribution"
					msg += "types, set discrete=False and grid_size=0."
					raise ValueError(msg)
				yvals.append(ydist.vals)
				yprobs.append(ydist.probs)
			return np.concatenate(yvals, axis=0), np.concatenate(yprobs, axis=0)


		### Continuous case ###
		if ninterp is None:
			ninterp = min(max(int(0.1 * (nvals-2)), 1), 5)
		# Case 2(a): we already have a discrete distribution
		if isinstance(ydists[0], utilities.BatchedCategorical):
			for ii, ydist in enumerate(ydists):
				if not isinstance(ydist, utilities.BatchedCategorical):
					raise ValueError(
						f"ydists[0] is a BatchedCategorical, ydists[{ii}] is not---mixing cts/discrete dists is unsupported."
					)
			# Extract and adjust everything to ensure the same support size
			nvals_adj = nvals - 2 * ninterp
			dists_adj = [
				utilities.adjust_support_size(
					vals=ydist.vals, probs=ydist.probs, new_nvals=nvals_adj, ymin=ymin, ymax=ymax
				)
				for ydist in ydists
			]
			yvals = np.concatenate([x[0] for x in dists_adj], axis=0)
			yprobs = np.concatenate([x[1] for x in dists_adj], axis=0)

		# Case 2(b): we have a continuous distribution
		else:
			if nvals <= MIN_NVALS:
				raise ValueError(f"nvals must be larger than {nvals}.")
			
			# make sure we get small enough quantiles
			alpha = min_quantile
			if alpha is None:
				alpha = 1 / (2*nvals)
			alpha = min(1/(2*nvals), max(alpha, 1e-8))

			# num of interp. pts between min/max quantiles
			# and ymin/ymax, added to ensure feasbility
			nvals_adj = nvals - 2 * ninterp - 3

			### Main discretization based on evenly spaced quantiles
			# choose endpoints of bins for disc. approx
			endpoints = np.sort(np.concatenate(
				[[0, alpha],
				np.linspace(1/nvals_adj, (nvals_adj-1)/nvals_adj, nvals_adj),
				[1-alpha, 1]],
			))
			qs = (endpoints[1:] + endpoints[0:-1])/2
			# loop through batches and concatenate
			yvals = []
			for dists in ydists:
				yvals.append(dists.ppf(qs.reshape(-1, 1)).T)
			yvals = np.concatenate(yvals, axis=0)
			n = len(yvals)
			# concatenate y probs
			yprobs = endpoints[1:] - endpoints[0:-1]
			yprobs = np.stack([yprobs for _ in range(n)], axis=0)

		### Insert additional support points with zero prob
		if ninterp > 0:
			to_add = np.linspace(ymin, yvals.min(axis=1), ninterp+1)[0:-1].T
			to_add = np.concatenate(
				[to_add, np.linspace(ymax, yvals.max(axis=1), ninterp+1)[0:-1].T],
				axis=1
			)
			yvals = np.concatenate([to_add, yvals], axis=1)
			yprobs = np.concatenate([np.zeros(to_add.shape), yprobs], axis=1)
	
		# Minimum yprob for numerical stability; then return
		yprobs = np.maximum(yprobs, min_yprob)
		yprobs /= yprobs.sum(axis=1).reshape(-1, 1)
		return yvals, yprobs

	def _ensure_feasibility(
		self,
		i,
		nu0,
		nu1,
		lower,
		y0_min,
		y0_max,
		y1_min,
		y1_max,
		grid_size=100,
		tol=5e-4,
	):
		"""
		ensures nu0 + nu1 <= fvals (if lower)
		or nu0 + nu1 >= fvals (if upper)
		by performing a gridsearch.

		Parameters
		----------
		i : int
			Index of which data-point we are performing
			this operation for.
		nu0 : np.array
			nvals0-length array of dual variables 
			associated with Y(0)
		nu1 : np.array
			nvals1-length array of dual variables
			associated with Y(1)
		lower : bool
			Specifies lower vs. upper bound.
		ymin : float
			Minimum value of y
		ymax : float
			Maximum value of y
		grid_size : int
			Grid size along each dimension (y(0) and y(1)).

		Returns
		-------
		new_y0_vals : np.array
			``nvals0 + grid_size`` length array of new y0_vals.
			*Exact size may change to avoid duplicate values
		new_nu0 : np.array
			``nvals0 + grid_size`` length array of interpolated
			and feasible dual vars. for Y(0).
		new_y1_vals : np.array
			``nvals1 + grid_size`` length array of new y1_vals
		new_nu1 : np.array
			``nvals1 + grid_size`` length array of interpolated
			and feasible dual vars. for Y(1).
			*Exact size may change to avoid duplicate values.
		dx : float
			Maximum numerical error induced by interpolation
			process.
		objval_diff : np.array
			The estimated objective value change for the new
			dual variables.
		"""
		obj_orig = nu0 @ self.y0_probs[i] + nu1 @ self.y1_probs[i]
		if grid_size > 0:
			# interpolate to compute new dual variables
			new_y0_vals = np.unique(np.sort(np.concatenate([
				np.linspace(y0_min, y0_max, grid_size), self.y0_vals[i]
			])))
			new_y1_vals = np.unique(np.sort(np.concatenate([
				np.linspace(y1_min, y1_max, grid_size), self.y1_vals[i]
			])))
			new_nu0 = self.interp_fn(
				x=self.y0_vals[i], y=nu0, newx=new_y0_vals,
			)
			new_nu1 = self.interp_fn(
				x=self.y1_vals[i], y=nu1, newx=new_y1_vals,
			)
			# compute required bounds
			fvals = self._apply_ot_fn(
				fn=self.f,
				y0=new_y0_vals,
				y1=new_y1_vals,
				x=self.X[i],
			)
			# support restrictions relax constraints
			if self.support_restriction is not None:
				not_in_support = ~(self._apply_ot_fn(
					fn=self.support_restriction,
					y0=new_y0_vals,
					y1=new_y1_vals,
					x=self.X[i],
				).astype(bool))
				# Relax constraints
				#almost_inf = 1e5 * (np.abs(fvals) + 1).max()
				if lower:
					fvals[not_in_support] = np.inf
				else:
					fvals[not_in_support] = - np.inf


			## Adjust elementwise (as opposed to a constant subtraction)
			# Calculate how far we are from feasibility
			deltas = new_nu0.reshape(-1, 1) + new_nu1.reshape(1, -1)
			deltas = deltas - fvals
			if lower:
				deltas0 = deltas.max(axis=1)
				dx0 = np.mean(np.maximum(deltas0, 0))
				deltas1 = deltas.max(axis=0)
				dx1 = np.mean(np.maximum(deltas1, 0))
				adj_axis = 1 if dx1 <= dx0 else 0
			else:
				deltas0 = deltas.min(axis=1)
				dx0 = np.mean(np.minimum(deltas0, 0))
				deltas1 = deltas.min(axis=0)
				dx1 = np.mean(np.minimum(deltas1, 0))
				adj_axis = 1 if dx1 >= dx0 else 0

			# Adjust and recompute interp_nu0/interp_nu1
			if adj_axis == 0:
				new_nu0 -= deltas0
				adj_nu0 = self.interp_fn(
					x=new_y0_vals, y=new_nu0, newx=self.y0_vals[i],
				)
				adj_nu1 = nu1
			else:
				new_nu1 -= deltas1
				adj_nu1 = self.interp_fn(
					x=new_y1_vals, y=new_nu1, newx=self.y1_vals[i],
				)
				adj_nu0 = nu0

			## Track change in objective
			new_objval = adj_nu0 @ self.y0_probs[i] + adj_nu1 @ self.y1_probs[i]
			objval_diff = obj_orig - new_objval

		else:
			new_y0_vals = self.y0_vals[i]
			new_nu0 = nu0
			new_y1_vals = self.y1_vals[i]
			new_nu1 = nu1
			objval_diff = 0

		# return
		return new_y0_vals, new_nu0, new_y1_vals, new_nu1, objval_diff

	def _ensure_primal_feasibility(
		self,
		y0_probs: np.array,
		y1_probs: np.array,
		y0_vals: np.array,
		y1_vals: np.array,
		not_in_support: np.array,
		i: int,
		fuzz=0.0001,
		solver='SCIPY',
		min_prob=0, ## new
	):
		"""
		fuzz : float
			Adds fuzz * np.random.uniform() to nonzero values to push 
			jointprobs into the interior of the primal feasible set.
		"""
		nvals0 = len(y0_vals)
		nvals1 = len(y1_vals)
		min_ratio = cp.Parameter()
		# Create variable
		jointprobs = cp.Variable((nvals0, nvals1), pos=True)
		y0p = jointprobs.sum(axis=1)
		y1p = jointprobs.sum(axis=0)
		constraints = [
			cp.sum(jointprobs)==1,
			jointprobs[not_in_support] == 0,
			y0p >= y0_probs * min_ratio,
			y1p >= y1_probs * min_ratio,
		]
		diffs0 = y0_vals[1:] - y0_vals[:-1]
		diffs1 = y1_vals[1:] - y1_vals[:-1]
		obj = cp.abs(cp.cumsum(y0p) - np.cumsum(y0_probs))[:-1] @ diffs0
		obj += cp.abs(cp.cumsum(y1p) - np.cumsum(y1_probs))[:-1] @ diffs1
		problem = cp.Problem(cp.Minimize(obj), constraints)

		for mr in [1/2, 1/5, 1/100, 0]:
			min_ratio.value = mr
			problem.solve(solver=solver)
			if problem.status in ['optimal', 'optimal_inaccurate']:
				break

		if problem.status not in ['optimal', 'optimal_inaccurate']:
			raise ValueError(f"For i={i}, failed to enforce primal feasibility.")
		# Clip
		jp = jointprobs.value
		jp += fuzz * np.random.uniform(0, 1, size=jp.shape) / (nvals0 * nvals1)
		jp[not_in_support] = 0
		jp = jp / jp.sum()
		return jp.sum(axis=1), jp.sum(axis=0)

	def _solve_single_instance(
		self, 
		i: int,
		probs0: np.array,
		probs1: np.array,
		y0_vals: np.array,
		y1_vals: np.array,
		fvals: np.array,
		not_binding: np.array,
		lower: bool,
		dual_strategy='ot',
		lp_solver='SCIPY',
		qp_solver='CLARABEL',
		se_solver='CLARABEL',
		enforce_primal_feas=True,
		debug=False,
		**kwargs,
	):
		"""
		Parameters
		----------
		not_binding : boolean array
			not_binding[i,j] = 1 iff the (i,j)th
			dual constraint should be ignored.
		debug : bool
			If True, becomes highly verbose.
		dual_strategy : str
			One of 'ot', 'lp', 'qp', 'se'
		"""
		if not lower:
			nu0, nu1, objval = self._solve_single_instance(
				probs0=probs0, 
				probs1=probs1,
				y0_vals=y0_vals,
				y1_vals=y1_vals,
				fvals=-1*fvals,
				i=i,
				not_binding=not_binding,
				lower=1,
				dual_strategy=dual_strategy,
				lp_solver=lp_solver,
				qp_solver=qp_solver,
				se_solver=se_solver,
				debug=debug,
				enforce_primal_feas=enforce_primal_feas,
				**kwargs,
			)
			return -nu0, -nu1, -objval

		### Parse dual solver
		dual_strategy = str(dual_strategy).lower()
		if dual_strategy not in ['ot', 'lp', 'qp', 'se']:
			raise ValueError(
				f"dual_strategy={dual_strategy} must be one of 'ot', 'lp', 'qp', 'se'"
			)

		### Structure
		# 1. initial OT solve  [used by ot]
		# 2. Ensure_primal_feas [used by all if any constraints are not binding]
		# 3. OT solve [used by ot]
		# 4. LP solve [used by lp]
		# 5. QP solve [used by qp]

		### 1. Initial OT solve
		if dual_strategy == 'ot':
			# Use a hack for non-binding constraints.
			# This only influences the OT solver.
			almost_inf = 1e5 * np.abs(fvals).max()
			if np.any(not_binding):
				fvals = fvals.copy()
				fvals[not_binding] = almost_inf

			# Attempt to solve with pyot
			objval, log = ot.lp.emd2(
				a=probs0,
				b=probs1,
				M=fvals,
				log=True,
				**kwargs,
			)
			nu0, nu1 = log['u'], log['v']
			too_large = _dualvals_are_too_large(nu0, nu1, almost_inf)
			if (not too_large) or (not np.any(not_binding)):
				nu1 += nu0.mean()
				nu0 -= nu0.mean()
				return nu0, nu1, objval

		### 2. Ensure primal feasibility
		if np.any(not_binding) and enforce_primal_feas:
			if debug:
				print(f"Ensuring primal feasibility. enforce_primal_feas={enforce_primal_feas}")
			probs0, probs1 = self._ensure_primal_feasibility(
				y0_probs=probs0,
				y1_probs=probs1, 
				y0_vals=y0_vals, 
				y1_vals=y1_vals, 
				not_in_support=not_binding, 
				i=i,
			)

		## 3. Second OT solve
		if dual_strategy == 'ot':
			if debug:
				print("Trying secondary ot.lp optimization.")
			objval, log = ot.lp.emd2(
				a=probs0,
				b=probs1,
				M=fvals,
				log=True,
				**kwargs,
			)
			nu0, nu1 = log['u'], log['v'] 
			nu1 += nu0.mean()
			nu0 -= nu0.mean()
			if _dualvals_are_too_large(nu0, nu1, almost_inf):
				dual_strategy = 'se'
			
		## 4/5: solve with LP/QP.
		if dual_strategy in ['lp', 'qp', 'se']:

			###NOTE: this is 25% faster but less general bc we can't swap in
			# new solvers.
			# c = np.concatenate([y0_probs, y1_probs])
			# i0s, i1s = np.where(in_support)
			# col_inds = np.concatenate([i0s, i1s + nvals0])
			# row_inds = np.concatenate([np.arange(len(i0s)), np.arange(len(i0s))])
			# b = fvals[(i0s, i1s)]
			# A_sparse = scipy.sparse.csr_matrix(
			#     (np.ones(len(row_inds)), (row_inds, col_inds)), shape=(len(i0s), len(c))
			# )
			# out_upper = scipy.optimize.linprog(
			#     c=-c, A_ub=A_sparse, b_ub=-b, bounds=(None, None),
			# )

			nu0 = cp.Variable(len(probs0))
			nu1 = cp.Variable(len(probs1))
			nusum = cp.reshape(nu0, (-1, 1)) + cp.reshape(nu1, (1, -1))
			linobj = nu0 @ probs0 + nu1 @ probs1
			binding = (~not_binding).astype(int)
			constraints = [cp.multiply(binding, nusum) <= cp.multiply(binding, fvals)]
			if dual_strategy == 'lp':
				if debug:
					print(f"Solving linear problem with {lp_solver} (not PyOT).")
				problem = cp.Problem(cp.Maximize(linobj), constraints=constraints)
				problem.solve(solver=lp_solver)
				if debug:
					print(f"LP problem status is {problem.status}.")
			elif dual_strategy == 'qp':
				if debug:
					print(f"Solving quadratic program with {qp_solver} (not PyOT).")
				pi = self.pis[i]
				pcat = np.maximum(
					np.concatenate([probs0 / (1-pi), probs1 / pi]), 
					1e-6
				)
				quad_term = (pcat @ cp.hstack([nu0, nu1]))**2 / self.n
				problem = cp.Problem(
					cp.Maximize(linobj - quad_term), constraints=constraints
				)
				problem.solve(solver=qp_solver)

			# maximize mean - 2 standard errors
			else:
				if debug:
					print(f"Solving se-adjusted problem with {se_solver} (not PyOT).")
				pcat = np.concatenate([probs0, probs1])
				pi = self.pis[i]
				Q = np.diag(np.concatenate([probs0 / (1-pi), probs1 / pi])) - np.outer(pcat, pcat)
				evals, evecs = np.linalg.eigh(Q)
				sqrtQ = evecs @ np.diag(np.sqrt(np.maximum(evals, 0))) @ evecs.T
				stdev = cp.norm2(sqrtQ @ cp.hstack([nu0, nu1]))
				problem = cp.Problem(
					cp.Maximize(linobj - 2 * stdev / np.sqrt(self.n)), constraints=constraints
				)
				problem.solve(solver=se_solver)

			# Warn the user of failures
			if problem.status not in ['optimal', 'optimal_inaccurate']:
				msg = "The final bounds may be vacuous."
				if problem.status == 'unbounded':
					msg = f"Dual solver failed due to primal infeasibility for i={i}, lower={lower}. " + msg
				else:
					msg = f"Dual solver failed for i={i}, lower={lower} with status {problem.status}. " + msg
				warnings.warn(msg)
				nu0 = np.zeros(len(probs0))
				nu1 = np.zeros(len(probs1))
				objval = 0
			else:
				nu0 = nu0.value
				nu1 = nu1.value
				objval = linobj.value

		# center
		if dual_strategy not in ['qp', 'se']:
			nu1 += nu0.mean()
			nu0 -= nu0.mean()
		return nu0, nu1, objval

	def compute_dual_variables(
		self,
		y0_dists: Optional[list]=None,
		y0_vals: Optional[np.array]=None,
		y0_probs: Optional[np.array]=None,
		y1_dists: Optional[list]=None,
		y1_vals: Optional[np.array]=None,
		y1_probs: Optional[np.array]=None,
		verbose: bool=True,
		min_quantile: Optional[float]=None,
		ninterp: Optional[int]=None,
		dual_strategy: str='ot',
		lp_solver: str='SCIPY',
		qp_solver: str='CLARABEL',
		se_solver: str='CLARABEL',
		nvals0: int=100,
		nvals1: int=100,
		interp_fn: callable=interpolation.adaptive_interpolate,
		y0_min: Optional[float]=None,
		y0_max: Optional[float]=None,
		y1_min: Optional[float]=None,
		y1_max: Optional[float]=None,
		**kwargs,
	):
		"""
		Estimates dual variables using the outcome model.

		We generally recommend that the user call .fit() instead of 
		calling this function directly.

		Parameters
		----------
		y0_dists : list
			The ith distribution of y0_dists represents the conditional
			law of :math:`Y_i(0) | X_i`. There are two input formats:

			- batched scipy distribution of shape (n,)
			- list of scipy dists whose shapes add up to n.

		y0_vals : list
			Alternatively, specify a (n, nvals0)-length array
			where ``y0_vals[i]`` is the support of :math:`Y_i(0)`.
			Ignored if ``y0_dists`` is provided.
		y0_probs : np.array
			A (n, nvals0)-length array where ``y0_probs[i, j]``
			is the estimated probability that :math:`Y_i(0)`
			equals ``y0_vals[i, j].``
		y1_dists : list
			The ith distribution of y1_dists represents the conditional
			law of :math:`Y_i(1) | X_i`. There are two input formats:

			- batched scipy distribution of shape (n,)
			- list of scipy dists whose shapes add up to n.

		y1_vals : np.array
			(n, nvals1)-length array where ``y1_vals[i]`` is the support 
			of :math:`Y_i(1)`.	Ignored if ``y1_dists`` is provided.
		y1_probs : np.array
			(n, nvals1)-length array where ``y1_probs[i, j]``
			is the estimated probability that :math:`Y_i(1)`
			equals ``y1_vals[i, j].``
		dual_strategy : str
			Specifies the strategy used to find the dual variables.
			One of the following:

			- 'ot': solves a standard-form optimal transport problem.
			- 'lp': solves a full linear program.
			- 'qp': solves a quadratic program which approximately accounts for standard errors.
			- 'se': solves a convex program which exactly accounts for standard errors.

			'ot' is the default and the fastest, but 'se' can reduce 
			standard errors in noisy problems.

		lp_solver : str
			When ``dual_strategy='lp'``, specifies which cvxpy solver 
			to use to solve the optimal transport LP. Default: SCIPY.
		qp_solver : str
			When ``dual_strategy='qp'``, specifies which cvxpy solver 
			to use to solve the optimal transport QP. Default: CLARABEL.
		se_solver : str
			When ``dual_strategy='se'``, specifies which cvxpy solver 
			to use to solve the convex program. Default: CLARABEL.
		min_quantile : float
			Minimum quantile to consider when discretizing. 
			Defaults to 1 / (2*nvals).		
		nvals0 : int
			How many values to use to discretize Y(0). 
			Defaults to 100. Ignored for discrete Y.
		nvals1 : int
			How many values to use to discretize Y(1).
			Defaults to 100. Ignored for discrete Y.
		interp_fn : function 
			An interpolation function with the same input/output
			signature as ``interpolation.adaptive_interpolate``,
			which is the default. Ignored for discrete Y.
		y0_min : float
			Minimum support for Y(0).
			Defaults to ``self.y.min() - 0.5 * (self.y.max() - self.y.min())``
		y1_min : float
			Minimum support for Y(1). 
			Defaults to ``self.y.min() - 0.5 * (self.y.max() - self.y.min())``
		y0_max : float
			Maximum support for Y(0). 
			Defaults to ``self.y.max() + 0.5 * (self.y.max() - self.y.min())``
		y1_max : float
			Maximum support for Y(1). 
			Defaults to ``self.y.max() + 0.5 * (self.y.max() - self.y.min())``
		kwargs : dict
			kwargs for ``_ensure_feasibility`` method, e.g., ``grid_size``.

		Returns
		-------
		None
		"""
		### Key quantities for optimizer
		# to ensure numerical stability, we add extra quantiles
		if verbose:
			print("Estimating optimal dual variables.")
		if min([nvals0, nvals1]) <= MIN_NVALS:
			raise ValueError(f"nvals0={nvals0}, nvals1={nvals1} must be larger than {MIN_NVALS}")
		
		# If discrete=True and ydists are BatchedCategorical distributions,
		# these values are ignored.
		self.nvals0 = nvals0
		self.nvals1 = nvals1
		self.interp_fn = interp_fn

		# max/min yvals
		y0_min, y1_min, y0_max, y1_max = _default_ylims(
			self.y, y0_min, y1_min, y0_max=y0_max, y1_max=y1_max
		)

		# this discretizes if Y is continuous
		if y0_vals is None or y0_probs is None:
			y0_vals, y0_probs = self._discretize(
				y0_dists, nvals=self.nvals0, min_quantile=min_quantile,
				ymin=y0_min, ymax=y0_max, ninterp=ninterp,
			)
			if self.discrete:
				self.nvals0 = y0_vals.shape[1]
		if y1_vals is None or y1_probs is None:
			y1_vals, y1_probs = self._discretize(
				y1_dists, nvals=self.nvals1, min_quantile=min_quantile,
				ymin=y1_min, ymax=y1_max, ninterp=ninterp,
			)
			if self.discrete:
				self.nvals1 = y1_vals.shape[1]

		# ensure y1_vals, y1_probs are sorted
		self.y0_vals, self.y0_probs = utilities._sort_disc_dist(y0_vals, probs=y0_probs)
		self.y1_vals, self.y1_probs = utilities._sort_disc_dist(y1_vals, probs=y1_probs)
		# Delete old values
		del y0_vals, y1_vals, y0_probs, y1_probs

		# Initialize results
		self.nu0s = np.zeros((2, self.n, self.nvals0)) # first dimension = [lower, upper]
		self.nu1s = np.zeros((2, self.n, self.nvals1))
		# Realized dual variables
		self.hatnu0s = np.zeros((2, self.n))
		self.hatnu1s = np.zeros((2, self.n))
		# estimated cond means of nu0s, nu1s
		self.c0s = np.zeros((2, self.n))
		self.c1s = np.zeros((2, self.n))
		# objective values
		self.objvals = np.zeros((2, self.n))
		self.dxs = np.zeros((2, self.n)) # ignored, only for backward compatability
		self.objdiffs = np.zeros((2, self.n))
		# loop through
		for i in utilities.vrange(self.n, verbose=verbose):
			# set parameter values
			fvals = self._apply_ot_fn(
				fn=self.f,
				y0=self.y0_vals[i],
				y1=self.y1_vals[i],
				x=self.X[i],
			)
			if self.support_restriction is not None:
				not_binding = ~(self._apply_ot_fn(
					fn=self.support_restriction,
					y0=self.y0_vals[i],
					y1=self.y1_vals[i],
					x=self.X[i],
				).astype(bool))
			else:
				not_binding = np.zeros(fvals.shape).astype(bool)
			# solve
			for lower in [1, 0]:
				nu0x, nu1x, objval = self._solve_single_instance(
					probs0=self.y0_probs[i],
					probs1=self.y1_probs[i],
					y0_vals=self.y0_vals[i],
					y1_vals=self.y1_vals[i],
					fvals=fvals,
					lower=lower,
					i=i,
					not_binding=not_binding,
					dual_strategy=dual_strategy,
					lp_solver=lp_solver,
					qp_solver=qp_solver,
					se_solver=se_solver,
				)
				self.objvals[1 - lower, i] = objval
				# Save solutions
				self.nu0s[1 - lower, i] = nu0x
				self.nu1s[1 - lower, i] = nu1x
				self.c0s[1 - lower, i] = nu0x @ self.y0_probs[i]
				self.c1s[1 - lower, i] = nu1x @ self.y1_probs[i]
				# Compute realized dual variables
				self._interpolate_and_ensure_feas(
					yi=self.y[i], 
					i=i,
					lower=lower, 
					y0_min=y0_min, 
					y0_max=y0_max,
					y1_min=y1_min,
					y1_max=y1_max,
					**kwargs
				)

	def _interpolate_and_ensure_feas(
		self, yi, lower, i, y0_min, y1_min, y0_max, y1_max, **kwargs
	):  
		"""
		This is used in two places:
		1. _compute_realized_dual_variables
		2. Main loop of compute_dual_variables

		We use this in the main loop of compute_dual_variables
		instead of just calling _compute_realized_dual_variables
		so that the progress report is accurate.
		"""
		nu0x = self.nu0s[1-lower, i]
		nu1x = self.nu1s[1-lower, i]
		# Ensure feasibility
		if not self.discrete:
			y0v, nu0adj, y1v, nu1adj, odx = self._ensure_feasibility(
				i=i, nu0=nu0x, nu1=nu1x, lower=lower, 
				y0_min=y0_min, y0_max=y0_max,
				y1_min=y1_min, y1_max=y1_max, 
				**kwargs
			)
			self.objdiffs[1-lower, i] = odx
			# interpolate to find realized values 
			self.hatnu0s[1 - lower, i] = self.interp_fn(
				x=y0v, y=nu0adj, newx=yi,
			)[0]
			self.hatnu1s[1 - lower, i] = self.interp_fn(
				x=y1v, y=nu1adj, newx=yi,
			)[0]
		else:
			j0 = np.argmin(np.abs(self.y0_vals[i] - yi))#.item()
			j1 = np.argmin(np.abs(self.y1_vals[i] - yi))#.item()
			self.hatnu0s[1 - lower, i] = nu0x[j0]
			self.hatnu1s[1 - lower, i] = nu1x[j1]   



	def _compute_realized_dual_variables(self, y=None, **kwargs):
		"""
		Helper function which applies interpolation 
		(for continuous Y) to compute realized dual variables.
		It also ensures feasibility through a 2D gridsearch.
		This is used primarily in unit tests.
		"""
		y = self.y if y is None else y

		# Create ylims
		y0_min, y1_min, y0_max, y1_max = _default_ylims(y=y, **kwargs)
		kwargs['y0_min'] = y0_min
		kwargs['y1_min'] = y1_min
		kwargs['y0_max'] = y0_max
		kwargs['y1_max'] = y1_max

		### Compute realized hatnu1s/hatnu0s
		for i in range(self.n):
			for lower in [0, 1]:
				self._interpolate_and_ensure_feas(
					i=i,
					yi=y[i],
					lower=lower, 
					**kwargs
				)

	def _compute_ipw_summands(self):
		"""
		Helper method to compute (A)IPW estimator summands.
		Must be run after running ``self.compute_dual_variables``
		"""
		# initialize outputs
		self.ipw_summands = np.zeros((2, self.n))
		self.aipw_summands = np.zeros((2, self.n))

		# compute IPW summands and AIPW summands
		for lower in [0, 1]:
			# IPW
			self.ipw_summands[1-lower] = self.hatnu1s[1-lower] / self.pis
			self.ipw_summands[1-lower][self.W == 0] = (
				self.hatnu0s[1-lower] / (1 - self.pis)
			)[self.W == 0]
			# AIPW
			mus = self.c1s[1-lower] + self.c0s[1-lower]
			aipw1 = (self.hatnu1s[1-lower] - self.c1s[1-lower]) / self.pis + mus 
			aipw0 = (self.hatnu0s[1-lower] - self.c0s[1-lower]) / (1-self.pis) + mus
			self.aipw_summands[1-lower] = aipw1 * self.W + (1 - self.W) * aipw0

	def _compute_final_bounds(self, aipw=True, alpha=0.05):
		"""
		Computes final results from the estimated dual variables.
		"""
		self._compute_ipw_summands()
		self.estimates, self.ses, self.cis = utilities.compute_est_bounds(
			summands=self.aipw_summands if aipw else self.ipw_summands,
			alpha=alpha,
			clusters=self.clusters,
		)
		return dict(
			estimates=self.estimates,
			ses=self.ses,
			cis=self.cis,
		)

	def _compute_cond_means(self):
		"""
		Helper function which computes self.mu0 = E[Y(0) | X],
		self.mu1 = E[Y(1) | X]. Can only be run after the 
		conditional distributions have been estimated.
		"""
		# make lists
		if isinstance(self.y0_dists, list):
			y0_dists = self.y0_dists
		else:
			y0_dists = [self.y0_dists]
		if isinstance(self.y1_dists, list):
			y1_dists = self.y1_dists
		else:
			y1_dists = [self.y1_dists]
		# Compute
		self.mu0 = np.concatenate([x.mean() for x in y0_dists])
		self.mu1 = np.concatenate([x.mean() for x in y1_dists])

	def fit_propensity_scores(
		self, nfolds: int, clip: float=1e-2, verbose: bool=True,
	):
		"""
		Cross-fits the propensity scores.

		Parameters
		----------
		nfolds : int
			Number of folds.
		clip : float
			Clip propensity scores to be in [clip, 1-clip].
		verbose : bool
			If True, prints progress reports.

		Returns
		-------
		None
		"""
		# Parse model
		if self.propensity_model is None:
			self.propensity_model = 'ridge'
		if isinstance(self.propensity_model, str):
			model_cls = dist_reg.parse_model_type(
				self.propensity_model, discrete=True
			)
			self.propensity_model = model_cls()
		
		# Loop through
		if verbose:
			print("Fitting propensity scores.")
		starts, ends = dist_reg.create_folds(n=self.n, nfolds=nfolds)
		# loop through and fit
		self.propensity_model_fits = []
		self.pis = np.zeros(self.n)
		for ii in utilities.vrange(len(starts), verbose=verbose):
			start, end = starts[ii], ends[ii]
			# Pick out data from the other folds
			not_in_fold = [
				i for i in np.arange(self.n) if i < start or i >= end
			]
			# Fit 
			model_fit = copy.deepcopy(self.propensity_model)
			model_fit.fit(
				X=self.X[not_in_fold], y=self.W[not_in_fold]
			)
			self.propensity_model_fits.append(model_fit)
			# Predict out of sample
			self.pis[start:end] = model_fit.predict_proba(
				self.X[start:end]
			)[:, 1]

		# Clip
		self.pis = np.minimum(np.maximum(self.pis, clip), 1-clip)

	def cross_fit(
		self,
		nfolds: int=5,
		suppress_warning: bool=False,
		verbose: bool=True,
		weight_by_propensities: bool=False,
	):
		"""
		Cross-fits the outcome model.

		Parameters
		----------
		nfolds : int
			Number of folds to use in cross-fitting.
		suppress_warning : bool
			If True, suppresses a potential warning about cross-fitting.
		verbose : bool
			If True, prints progress reports.
		weight_by_propensities : bool
			If True, when cross-fitting the outcome model, upweights
			observations with low propensity scores.

		Returns
		-------
		y0_dists : list
			list of batched scipy distributions whose shapes sum to n.
			the ith distribution is the out-of-sample estimate of
			the conditional law of :math:`Y_i(0) | X[i]`
		y1_dists : list
			list of batched scipy distributions whose shapes sum to n.
			the ith distribution is the out-of-sample estimate of
			the conditional law of :math:`Y_i(1) | X[i]`
		"""

		# if pis not supplied: will use cross-fitting
		if self.pis is None:
			self.fit_propensity_scores(nfolds=nfolds, verbose=verbose)

		# Fit model
		if self.y0_dists is None or self.y1_dists is None:
			# Note: this returns the existing model
			# if an existing model is provided
			self.outcome_model = get_default_model(
				discrete=self.discrete, 
				support=self.support,
				outcome_model=self.outcome_model,
				**self.model_kwargs
			)
			if verbose:
				print("Cross-fitting the outcome model.")
			# Possibly create sample weights
			if weight_by_propensities:
				self.sample_weight = 1 / self.pis.copy()
				self.sample_weight[self.W == 0] = 1 / (1 - self.pis[self.W == 0])
			else:
				self.sample_weight = None

			y_out = dist_reg.cross_fit_predictions(
				W=self.W, X=self.X, y=self.y, 
				sample_weight=self.sample_weight, 
				nfolds=nfolds, 
				model=self.outcome_model,
				model_selector=self.model_selector,
				verbose=verbose,
			)
			counterfactuals, self.model_fits, self.oos_dist_preds = y_out
			self.y0_dists = counterfactuals[0]
			self.y1_dists = counterfactuals[1]
		elif not suppress_warning:
			warnings.warn(CROSSFIT_WARNING)

		return self.y0_dists, self.y1_dists

	def _compute_oos_resids(self):
		# compute out-of-sample predictions
		self._compute_cond_means()
		self.oos_preds = self.mu0.copy()
		self.oos_preds[self.W == 1] = self.mu1[self.W == 1]
		# residuals and return
		self.oos_resids = self.y - self.oos_preds
		return self.oos_resids

	def fit(
		self,
		nfolds: int = 5,
		aipw: bool = True,
		alpha: float = 0.05,
		y0_dists: Optional[list[rv_generic]] = None,
		y1_dists: Optional[list[rv_generic]] = None,
		verbose: bool = True,
		suppress_warning: bool = False,
		weight_by_propensities: bool = False,
		**solve_kwargs,
	):
		"""
		Main function which (1) performs cross-fitting, (2) computes 
		optimal dual variables, and (3) computes final dual bounds.

		Parameters
		----------  
		nfolds : int
			Number of folds to use when cross-fitting. Defaults to 5.
		alpha : float
			Nominal coverage level. Defaults to 0.05.
		aipw : bool
			If true, returns AIPW estimator.
		y0_dists : list
			The ith distribution of y0_dists represents the conditional
			law of :math:`Y_i(0) | X_i`. There are two input formats:

			- batched scipy distribution of shape (n,)
			- list of scipy dists whose shapes add up to n.

			This is an optional input; if provided, ``outcome_model``
			will be ignored.
		y1_dists : list
			The ith distribution of y1_dists represents the conditional
			law of :math:`Y_i(1) | X_i`. There are two input formats:

			- batched scipy distribution of shape (n,)
			- list of scipy dists whose shapes add up to n.

			This is an optional input; if provided, ``outcome_model``
			will be ignored.
		verbose : bool
			If True, gives occasional progress reports.
		suppress_warning : bool
			If True, suppresses a warning about cross-fitting.
		weight_by_propensities : bool
			If True, when cross-fitting the outcome model, upweights
			observations with low propensity scores.
		solve_kwargs : dict
			Additional (optional) kwargs for the ``compute_dual_variables``
			method, e.g. ``nvals0``, ``nvals1``, ``grid_size``.

		Returns
		-------
		self
		"""
		# Fit model of W | X and Y | X if not provided
		self.y0_dists, self.y1_dists = y0_dists, y1_dists
		self.cross_fit(
			nfolds=nfolds, 
			suppress_warning=suppress_warning, 
			verbose=verbose,
			weight_by_propensities=weight_by_propensities,
		)

		# compute dual variables
		self.compute_dual_variables(
			y0_dists=self.y0_dists,
			y1_dists=self.y1_dists,
			verbose=verbose,
			**solve_kwargs,
		)
		# compute dual bounds
		self.alpha = alpha
		self._compute_final_bounds(aipw=aipw, alpha=alpha)
		return self

	def _plug_in_results(self):
		pests, pses, pcis = utilities.compute_est_bounds(
			summands=self.objvals,
			alpha=self.alpha
		)
		return pd.DataFrame(
			np.stack(
				[pests, pses, pcis], 
				axis=0
			),
			index=['Estimate', 'SE', 'Conf. Int.'],
			columns=['Lower', 'Upper']
		)

	def results(self, minval: float=-np.inf, maxval: float=np.inf):
		"""
		Returns a dataframe of key inferential results.

		Parameters
		----------
		minval : float
			Analytical lower bound on estimand used to clip results. 
			Defaults to -np.inf.
		maxval : float
			Analytical upper bound on estimand used to clip results.
			Defaults to np.inf.

		Returns
		-------
		results : pd.DataFrame
			DataFrame of key inferential results.
		"""
		self.results_ = pd.DataFrame(
			np.stack(
				[
					np.clip(self.estimates, minval, maxval), 
					self.ses,
					np.clip(self.cis, minval, maxval),
				], 
				axis=0
			),
			index=['Estimate', 'SE', 'Conf. Int.'],
			columns=['Lower', 'Upper']
		)
		return self.results_

	def plot_dual_variables(self, i=0):
		"""
		Plots the estimated dual variables for the ith data-point.

		Parameters
		----------
		i : int
			Integer ranging from 0 to n-1, specifying which datapoint to plot.
		"""
		import matplotlib.pyplot as plt
		fig, axes = plt.subplots(1, 2, figsize=(10, 4))
		for name, j, ax in zip(['Lower', 'Upper'], [0,1], axes):
			# Plot dual variables
			for yvals, nus, color, label in zip(
				[self.y0_vals[i], self.y1_vals[i]],
				[self.nu0s[j, i], self.nu1s[j, i]],
				['red', 'blue'],
				['Control', 'Treatment'],
			):
				axes[j].scatter(yvals, nus, color=color, label=label)
			# Plot realized value
			hatnu = self.hatnu0s[j, i] if self.W[i] == 0 else self.hatnu1s[j, i]
			axes[j].axvline(self.y[i], color='black', label='Realized outcome value')
			axes[j].scatter(self.y[i], hatnu, color='red' if self.W[i] == 0 else 'blue')
			axes[j].set(xlabel='Outcome', ylabel='Dual variable', title=f'{name} Bound')
			axes[j].legend()
		plt.show()

	def diagnostics(self, plot=False, aipw=True):
		"""
		Reports a set of technical diagnostics.

		Parameters
		----------
		plot : bool
			If True, creates a set of diagnostic plots.

		Returns
		-------
		df : pd.DataFrame
			DataFrame of technical diagnostic information.

		Notes
		-----
		Please see the user guide for more details on the
		meaning of the outputs.
		"""
		summands = self.aipw_summands if aipw else self.ipw_summands
		aipw_name = 'AIPW' if aipw else 'IPW'
		# Plot AIPW dual summands
		if plot:
			import matplotlib.pyplot as plt
			fig, axes = plt.subplots(1, 2, figsize=(10, 4))
			for j, name in zip([0,1], ['Lower', 'Upper']):
				for w, color, label in zip([0,1], ['red', 'blue'], ['Control', 'Treatment']):
					axes[j].scatter(
						self.y[self.W == w], 
						self.aipw_summands[j][self.W == w], 
						color=color, 
						label=label
					)
				axes[j].legend()
				axes[j].set(xlabel='Outcome', ylabel=f'Dual {aipw_name} Summand', title=f"{name} Dual Bound")
			plt.show()
		# Leverage
		s2s = (summands - summands.mean(axis=1).reshape(-1, 1))**2
		leverages = np.max(s2s, axis=1) / s2s.sum(axis=1)
		# Max contribution
		max_contribs = np.array(
			[summands[0].min() / self.n, summands[1].max() / self.n]
		)
		return pd.DataFrame(
			np.stack(
				[self.objdiffs.mean(axis=1), leverages, max_contribs],
				axis=0
			),
			index=['Loss from gridsearch', 'Max leverage', f'Worst dual {aipw_name} summand'],
			columns=['Lower', 'Upper']
		)
		

	def eval_outcome_model(self):
		"""
		Thinly wraps ``dist_reg._evaluate_model_predictions``.

		Returns
		-------
		sumstats : pd.DataFrame
			DataFrame summarizing goodness-of-fit metrics for
			the cross-fit outcome model.
		"""
		self._compute_oos_resids()
		return dist_reg._evaluate_model_predictions(
			y=self.y, haty=self.oos_preds
		)

	def eval_treatment_model(self):
		"""
		Thinly wraps ``dist_reg._evaluate_model_predictions``.

		Returns
		-------
		sumstats : pd.DataFrame
			DataFrame summarizing goodness-of-fit metrics for
			the cross-fit propensity scores.
		"""
		return dist_reg._evaluate_model_predictions(
			y=self.W, haty=self.pis
		)

	def summary(self, minval: float=-np.inf, maxval: float=np.inf):
		"""
		Prints a summary of main results from the class.

		Parameters
		----------
		minval : float
			Analytical lower bound on estimand used to clip results. 
			Defaults to -np.inf.
		maxval : float
			Analytical upper bound on estimand used to clip results.
			Defaults to np.inf.

		Returns
		-------
		None

		Notes
		-----
		To access the results a DataFrame, call:

		- ``DualBounds.results()`` for inferential results
		- ``DualBounds.eval_outcome_model()`` for outcome model metrics
		- ``DualBounds.eval_treatment_model()`` for treatment model metrics
		- ``DualBounds._plug_in_results()`` for nonrobust plug-in bounds
		- ``DualBounds.diagnostics()`` for technical diagnostic information
		"""
		print("___________________Inference_____________________")
		print(self.results(minval=minval, maxval=maxval))
		print()
		print("_________________Outcome model___________________")
		print(self.eval_outcome_model())
		print()
		print("_________________Treatment model_________________")
		print(self.eval_treatment_model())
		print()
		print("______________Nonrobust plug-in bounds___________")
		print(self._plug_in_results())
		print()
		# possibly print diagnostics. This logic is useful
		# because some classes inheriting from 
		# DualBounds (e.g. VarCATEDualBounds) don't produce diagnostics.
		diagnostics = self.diagnostics(plot=False)
		if diagnostics is not None:
			print("_______________Technical diagnostics_____________")
			print(diagnostics)
			print()

def _plug_in_no_covariates(
	f: callable,
	y: np.array,
	W: np.array,
	pis: Optional[np.array]=None,
	max_nvals: int=1000,
) -> float:
	"""
	Internal function which does the work for plug_in_no_covariates
	and gets called in each bootstrap iteration.
	"""
	n = len(y)
	# Create empirical distributions for treatment/control
	y0_vals = y[W == 0]
	y0_probs = 1 / (1-pis[W==0]); y0_probs /= y0_probs.sum()
	y1_vals = y[W == 1]
	y1_probs = 1 / (pis[W==1]); y1_probs /= y1_probs.sum()
	# Reduce dimension to prevent memory errors for huge datasets
	qs = np.linspace(1/(max_nvals+1), max_nvals/(max_nvals+1), max_nvals)
	if len(y0_vals) > max_nvals:
		y0_vals = utilities.weighted_quantile(y0_vals, y0_probs, quantiles=qs)
		y0_probs = np.ones(len(y0_vals)) / len(y0_vals)
	if len(y1_vals) > max_nvals:
		y1_vals = utilities.weighted_quantile(y1_vals, y1_probs, quantiles=qs)
		y1_probs = np.ones(len(y1_vals)) / len(y1_vals)
	# Fvals
	fvals = f(y0_vals.reshape(-1, 1), y1_vals.reshape(1, -1), x=0)
	# lower and upper
	lower = ot.lp.emd2(a=y0_probs, b=y1_probs, M=fvals)
	upper = -ot.lp.emd2(a=y0_probs, b=y1_probs, M=-fvals)
	return np.array([lower, upper])

def plug_in_no_covariates(
	outcome: np.array, 
	treatment: np.array, 
	f: callable, 
	propensities: Optional[np.array]=None,
	clusters: Optional[np.array]=None,
	B: int=0,
	verbose: bool=True,
	alpha: float=0.05,
	max_nvals: int=1000,
	_which_bound='both',
) -> dict:
	"""
	Computes plug-in bounds on :math:`E[f(Y(0),Y(1))]` without using covariates.

	Parameters
	----------
	outcome : np.array
		n-length array of outcomes (y)
	treatment : np.array
		n-length array of treatments (W).
	f : function
		f(y0, y1, x) defines the objective.
	propensities : np.array
		n-length array of propensity scores (pis).
		Default: all equal to treatment.mean().
	clusters : np.array
		Optional n-length array of clusters, so ``clusters[i] = j``
		indicates that observation i is in cluster j.
	B : int
		Number of bootstrap replications to compute standard errors.
		Defaults to 0 (no standard errors).
	verbose : bool
		Show progress bar while bootstrapping if verbose=True.
	alpha : float
		nominal Type I error level.
	max_nvals : int
		Maximum dimension of optimal transport problem.
	_which_bound : str
		One of 'both', 'lower', 'upper'.

	Returns
	-------
	results : dict
		Dictionary containing up to three keys:

		- estimates: 2-length array of lower/upper estimates.
		- ses: 2-length array of lower/upper standard errors.
		- cis: 2-length array of lower/upper confidence intervals.

		These arrays will be length 1 (instead of 2) if which_bound != 'both'.
	"""
	# Infer propensities
	if propensities is None:
		propensities = np.ones(len(treatment)) * treatment.mean()
	# Create estimates
	estimates = _plug_in_no_covariates(
		f=f,
		y=outcome,
		W=treatment,
		pis=propensities,
		max_nvals=max_nvals
	)	
	if B == 0:
		return dict(estimates=estimates)

	# Compute bootstrapped SEs
	data = np.stack([outcome, treatment, propensities], axis=1)
	func = lambda data: _plug_in_no_covariates(
		f=f,
		y=data[:, 0],
		W=data[:, 1],
		pis=data[:, 2],
		max_nvals=max_nvals,
	)
	ses = utilities.cluster_bootstrap_se(
		data=data,
		clusters=clusters,
		func=func,
		B=B,
		verbose=verbose,
	)[0]
	cis = estimates.copy()
	cis[0] -= stats.norm.ppf(1-alpha/2) * ses[0]
	cis[1] += stats.norm.ppf(1-alpha/2) * ses[1]
	return dict(
		estimates=estimates,
		ses=ses,
		cis=cis,
	)


import setuptools
import codecs
import os.path

def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()

def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="dualbounds",
    version=get_version('dualbounds/__init__.py'),
    author="Asher Spector",
    author_email="amspector100@gmail.com",
    description="Dual bounds for model-agnostic inference",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/amspector100/dualbounds/",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.9',
    install_requires=[
        "numpy>=1.17.4",
        "scipy>=1.12.0",
        "scikit-learn>=1.4.1",
        "pandas",
        "POT",
        "cvxpy",
        "tqdm",
    ],
)


"""
Generate synthetic data for tests and illustrations.
"""

import numpy as np
from scipy import stats
from scipy.special import logsumexp
from .utilities import parse_dist, BatchedCategorical
from typing import Optional, Union

def heteroskedastic_scale(X, heterosked='constant'):
	n, p = X.shape
	if heterosked == 'constant' or heterosked == 'none':
		scale = np.ones(n)
	elif heterosked == 'linear':
		scale = np.abs(X[:, 0])
	elif heterosked == 'norm':
		scale = np.power(X, 2).sum(axis=-1)
	elif heterosked == 'invnorm':
		scale = 1 / (np.power(X, 2).sum(axis=-1))
	elif heterosked == 'exp_linear':
		scale = np.sqrt(np.exp(X[:, 0] + X[:, 1]))
	else:
		raise ValueError(f"Unrecognized heterosked={heterosked}")

	# normalize to ensure Var(epsilon) = 1 marginally
	scale /= np.sqrt(np.power(scale, 2).sum() / n)
	return scale

def create_cov(p, covmethod='identity'):
	covmethod = str(covmethod).lower()
	if covmethod == 'identity':
		return np.eye(p)
	else:
		raise ValueError(f"Unrecognized covmethod={covmethod}")

def _sample_norm_vector(dim, norm, sparsity=0):
	""" samples x ~ N(0, dim) and then normalizes so |x|_2 = norm """
	if norm == 0:
		return np.zeros(dim)
	# Possibly select sparse subset
	if sparsity > 0:
		x = np.zeros(dim)
		n_nonzero = int(max(1, np.ceil((1-sparsity) * dim)))
		x[np.random.choice(dim, n_nonzero, replace=False)] = _sample_norm_vector(
			dim=n_nonzero, norm=norm, sparsity=0,
		)
		return x
	# Else create a dense vector
	else:
		x = np.random.randn(dim)
		return x / np.sqrt(np.power(x, 2).sum() / norm)

def gen_regression_data(
	n: int,
	p: int,
	lmda_dist: str='constant',
	eps_dist: str='gaussian',
	heterosked: str='constant',
	tauv: float=1, # Var(Y(1) | X ) / Var(Y(0) | X)
	r2: float=0.95,
	sparsity: float=0,
	interactions: bool=True,
	tau: float=3,
	betaW_norm: float=0,
	covmethod: str='identity',
	dgp_seed: int=1,
	sample_seed: Optional[int]=None,
):
	"""
	Samples a synthetic regression dataset.

	Parameters
	----------
	n : int
		Number of observations.
	p : int
		Number of covariates
	lmda_dist : str
		str specifying the distribution of lmdai, 
		where Xi = lmdai * N(0, Sigma), so the 
		covariates are elliptically distributed. 
	eps_dist : str
		str specifying the distribution of the 
		residuals. See ``utilities.parse_dist``
		for the list of options.
	heterosked : str
		str specifying the type of heteroskedasticity.
		Defaults to ``constant``.
	tauv : float
		Ratio of Var(Y(1) | X) / Var(Y(0) | X)
	r2 : float
		Population r^2 of 1 - E[Var(Y | X)] / Var(Y).
	sparsity : float
		Proportion of covariates with zero coefficients.
		Defaults to zero (no sparsity).
	interactions : bool	
		If True (default), 
		Y = X beta + W * X * beta_int + epsilon.
		Else, the interactions between the treatment
		and the covariates are ommitted.
	tau : float
		Average treatment effect.
	betaW_norm : float
		E[W | X] = logistic(X @ betaW). This parameter
		controls the norm of betaW and thus the 
		variance of the propensity scores.
	covmethod : str
		str identifier for how to generate the covariance matrix.
	dgp_seed : int
		Random seed for the data-generating parameters.
	sample_seed : int
		Random seed for the randomness from sampling.

	Returns
	-------
	data : dict
		Dictionary with keys ``X`` (covariates),
		``y`` (response), ``W`` (treatment), 
		``pis`` (true propensity scores), 
		``beta``, ``beta_int``, ``betaW``, 
		and more.
	"""
	# create parameters
	np.random.seed(dgp_seed)
	Sigma = create_cov(p=p, covmethod=covmethod)
	# Create beta
	beta_norm = r2 / (1 - r2) if r2 > 0 else 0
	# No interactions with treatment
	if not interactions:
		beta = _sample_norm_vector(p, beta_norm, sparsity=sparsity)
		beta_int = np.zeros(p)
	# interactions with treatment
	else:
		beta = _sample_norm_vector(2*p, beta_norm, sparsity=sparsity)
		beta_int = beta[p:].copy()
		beta = beta[0:p].copy()

	# Create beta_W (for W | X)
	betaW = _sample_norm_vector(p, betaW_norm, sparsity=sparsity)

	# sample X
	np.random.seed(sample_seed)
	X = np.random.randn(n, p)
	L = np.linalg.cholesky(Sigma)
	X = X @ L.T
	lmdas = parse_dist(lmda_dist).rvs(size=n)
	lmdas /= np.sqrt(np.power(lmdas, 2).mean())
	X = X * lmdas.reshape(-1, 1)

	# sample W
	muW = X @ betaW
	pis = np.exp(muW)
	pis = pis / (1 + pis)
	# clip in truth
	pis = np.maximum(np.minimum(pis, 1-1e-3), 1e-3)
	W = np.random.binomial(1, pis)

	# conditional mean of Y
	mu = X @ beta
	# conditional variance of Y
	sigmas = heteroskedastic_scale(X, heterosked=heterosked)
	# allow for sigmas to depend on W
	sigmas0 = sigmas.copy()
	sigmas1 = tauv * sigmas.copy()
	denom = np.sqrt((np.mean(sigmas0**2) + np.mean(sigmas1**2)) / 2)
	sigmas0 /= denom; sigmas1 /= denom
	sigmas0 = np.maximum(1e-5, sigmas0)
	sigmas1 = np.maximum(1e-5, sigmas1)
	# Sample Y
	y0_dists = parse_dist(
		eps_dist, loc=mu, scale=sigmas0
	)
	cates = X @ beta_int + tau
	y1_dists = parse_dist(
		eps_dist, loc=mu+cates, scale=sigmas1
	)
	Y0 = y0_dists.rvs(); Y1 = y1_dists.rvs()
	Y = Y0.copy(); Y[W == 1] = Y1[W == 1]
	# return everything
	out = dict(
		X=X,
		W=W,
		y=Y,
		y0_dists=y0_dists,
		y1_dists=y1_dists,
		pis=pis,
		Sigma=Sigma,
		beta=beta,
		beta_int=beta_int,
		cates=y1_dists.mean() - y0_dists.mean(),
		betaW=betaW,
	)
	# # for convenience
	# if eps_dist == 'bernoulli':
	# 	out.update(dict(
	# 		_y0_dists_4input=BatchedCategorical.from_scipy(y0_dists),
	# 		_y1_dists_4input=BatchedCategorical.from_scipy(y1_dists),
	# 	))
	# else:
	# 	out.update(dict(
	# 		_y0_dists_4input=y0_dists,
	# 		_y1_dists_4input=y1_dists,
	# 	))
	return out

def gen_lee_bound_data(
	stau: float=1, betaS_norm: float=1, **kwargs
):
	"""
	Generates synthetic datasets with selection bias.

	Parameters
	----------
	stau : float
		In the logistic regression of S on W and X,
		stau is the coefficient on W.
	betaS_norm : float
		norm of coefficients in logistic regression of S 
		on W and X.
	"""
	# Generate regression data
	output = gen_regression_data(**kwargs)
	X, W = output['X'], output['W']
	n, p = X.shape
	# create DGP for S | X
	np.random.seed(kwargs.get("dgp_seed", 1))
	betaS = _sample_norm_vector(p, norm=betaS_norm)
	# sample S | X
	np.random.seed(kwargs.get("sample_seed", None))
	muS0 = X @ betaS
	muS1 = muS0 + stau
	s0_probs = np.exp(muS0) / (1 + np.exp(muS0))
	s1_probs = np.exp(muS1) / (1 + np.exp(muS1))
	S0 = np.random.binomial(1, s0_probs)
	S1 = np.random.binomial(1, s1_probs)
	S = S0.copy(); S[W == 1] = S1[W == 1]
	# save and return
	for key, val in zip(
		['s0_probs', 's1_probs', 'S'],
		[s0_probs, s1_probs, S]
	):
		output[key] = val
	return output

def gen_iv_data(
	n: int,
	p: int,
	betaZ_norm: float=0,
	betaW_norm: float=0.5,
	r2: float=0.95,
	sparsity: float=0,
	eps_dist: str='gaussian',
	lmda_dist: str='constant',
	interactions: bool=True,
	tau: float=3,
	tauv: float=1,
	tauZ: float=1,
	tau_conf: float=0,
	heterosked: str='constant',
	dgp_seed: int=1,
	sample_seed: Optional[int]=None,
):
	"""
	Generates IV data.
	
	Parameters
	----------
	n : int
		Number of observations.
	p : int
		Number of covariates
	lmda_dist : str
		str specifying the distribution of lmdai, 
		where Xi = lmdai * N(0, Sigma), so the 
		covariates are elliptically distributed. 
	eps_dist : str
		str specifying the distribution of the 
		residuals. See ``utilities.parse_dist``
		for the list of options.
	heterosked : str
		str specifying the type of heteroskedasticity.
		Defaults to ``constant``.
	tau : float
		Average treatment effect.
	tauv : float
		Ratio of Var(Y(1) | X) / Var(Y(0) | X)
	r2 : float
		Population r^2 of 1 - E[Var(Y | X)] / Var(Y).
	sparsity : float
		Proportion of covariates with zero coefficients.
		Defaults to zero (no sparsity).
	interactions : bool
		If True, there are interactions in the regression
		of Y on X.
	betaZ_norm: float
		Norm of beta in logistic regression of Z on X.
	betaW_norm : float
		Norm of beta in logistic regression of W on X, Z.
	tauZ : float
		Coefficient of Z in logistic regression of W on X, Z.
	dgp_seed : int
		Random seed for the data-generating parameters.
	sample_seed : int
		Random seed for the randomness from sampling.

	Returns
	-------
	data_dictionary : dict
		Dictionary of various data-generating parameters plus raw data.
	"""
	# create parameters
	np.random.seed(dgp_seed)
	# Create beta
	beta_norm = r2 / (1 - r2) if r2 > 0 else 0
	# No interactions with treatment
	if not interactions:
		beta = _sample_norm_vector(p, beta_norm, sparsity=sparsity)
		beta_int = np.zeros(p)
	# interactions with treatment
	else:
		beta = _sample_norm_vector(2*p, beta_norm, sparsity=sparsity)
		beta_int = beta[p:].copy()
		beta = beta[0:p].copy()

	# Create beta_W (for W | X)
	betaW = _sample_norm_vector(p, betaW_norm, sparsity=sparsity)
	betaZ = _sample_norm_vector(p, betaZ_norm, sparsity=sparsity)

	# sample X
	np.random.seed(sample_seed)
	X = np.random.randn(n, p)
	lmdas = parse_dist(lmda_dist).rvs(size=n)
	lmdas /= np.sqrt(np.power(lmdas, 2).mean())
	X = X * lmdas.reshape(-1, 1)

	### Sample Z | X
	muZ = X @ betaZ
	pis = np.exp(muZ)
	pis = pis / (1 + pis)
	# clip in truth
	pis = np.maximum(np.minimum(pis, 1-1e-3), 1e-3)
	Z = np.random.binomial(1, pis).astype(int)

	### Sample W | Z, X. Here U are meant to be unmeasured.
	wprobs = np.stack([X @ betaW, X @ betaW + tauZ], axis=1)
	wprobs = np.exp(wprobs) / (1 + np.exp(wprobs))
	U = np.random.uniform(size=n)
	W01 = (U.reshape(n, 1) <= wprobs).astype(int)
	W = W01[(np.arange(n), Z)]

	### Sample Y | Z, W, X
	## (a) conditional variance of Y
	sigmas = heteroskedastic_scale(X, heterosked=heterosked)
	# allow for sigmas to depend on W
	sigmas0 = sigmas.copy()
	sigmas1 = tauv * sigmas.copy()
	denom = np.sqrt((np.mean(sigmas0**2) + np.mean(sigmas1**2)) / 2)
	sigmas0 /= denom; sigmas1 /= denom
	## (b) conditional mean of Y | X, W U
	mu0 = X @ beta + tau_conf * np.sign(U - 1/2)
	_y0_dists = parse_dist(
		eps_dist, mu=mu0, sd=sigmas0
	)
	cates = X @ beta_int + tau
	mu1 = mu0 + cates
	_y1_dists = parse_dist(
		eps_dist, mu=mu1, sd=sigmas1
	)
	Y0 = _y0_dists.rvs(); Y1 = _y1_dists.rvs()
	Y = Y0.copy(); Y[W == 1] = Y1[W == 1]
	## (c) ydists for external use. 
	# ydists[z][w] are the laws of Y(w) | W(z) = w
	ydists = [[], []]
	if tau_conf != 0:
		warnings.warn("Oracle ydists not available when tau_conf != 0; need to implement ScipyMixture.")
	else:
		for z in [0,1]:
			for w, muY, sigmaY in zip([0,1], [mu0, mu1], [sigmas0, sigmas1]):
				ydists[z].append(parse_dist(eps_dist, mu=muY, sd=sigmaY))

	return dict(
		y=Y,
		W=W,
		Z=Z,
		X=X,
		pis=pis,
		cates=_y1_dists.mean() - _y0_dists.mean(),
		# betas
		beta=beta,
		beta_int=beta_int,
		betaZ=betaZ,
		betaW=betaW,
		# distributions
		wprobs=wprobs,
		ydists=ydists,
	)


# class ScipyMixture:
# 	"""
# 	A finite mixture of scipy distributions.

	
# 	Parameters
# 	---------
# 	distributions : rv_generic
# 		A list of scipy distributions of shape (n,).
# 	proportions : np.array
# 		The mixture proportions. Must sum to 1.

# 	Notes
# 	-----
# 	This is only used within the ``gen_iv_data``
# 	function to calculate the law of Y(w) | W(z) = w
# 	when tau_conf != 0.
# 	"""
# 	def __init__(
# 		self, distributions, proportions
# 	):
# 		self.dists = distributions
# 		self.props = proportions
# 		if not np.allclose(np.sum(self.props), 1):
# 			raise ValueError("proportions must sum to 1")

# 	def mean(self):
# 		"""
# 		Returns
# 		-------
# 		mu : np.array
# 			n-shaped array of means of mixture distribution.
# 		"""
# 		mus = np.stack([d.mean() * p for d, p in zip(self.dists, self.props)], axis=0)
# 		return np.sum(mus, axis=0)

# 	def cdf(self, x):
# 		"""
# 		Parameters
# 		----------
# 		x : np.array
# 			Inputs to CDF.

# 		Returns
# 		-------
# 		cdf : np.array
# 			CDF realized values.
# 		"""
# 		cdfs = np.stack([d.cdf() * p for d, p in zip(self.dists, self.props)], axis=0)
# 		return np.sum(cdfs, axis=0)

# 	def ppf(self, qs):
# 		"""
# 		Parameters
# 		----------
# 		qs : np.array
# 			Inputs to quantile function.

# 		Returns
# 		-------
# 		vals : np.array
# 			Quantile function evaluated at qs.
# 		"""
# 		pass


import copy
import warnings
import numpy as np
import ot
import cvxpy as cp
from scipy import stats
from . import utilities, generic, dist_reg, interpolation
from .utilities import BatchedCategorical
import pandas as pd
# typing
from scipy.stats._distn_infrastructure import rv_generic
import sklearn.base
from typing import Optional, Union

"""
Note to self: do we need to define a ConcatScipyDist wrapper...?
"""

def _clip_jointprobs(jointprobs):
	jointprobs = np.clip(jointprobs, 0, 1)
	for z in [0,1]:
		jointprobs[z] /= jointprobs[z].sum()
	return jointprobs

def _compute_max_dualval(logs):
	max_dualval = 0
	for lower in [0,1]:
		for key in ['u', 'v']:
			max_dualval = max(np.abs(logs[lower][key]).max(), max_dualval)
	return max_dualval


class DualIVBounds(generic.DualBounds):
	"""
	Beta version. Computes dual bounds on 
	:math:`E[f(W(0), W(1), Y(0),Y(1), X)].`
	in the instrumental variables context.

	Here, :math:`X` are covariates, :math:`Y(0), Y(1)` are 
	potential outcomes, and :math:`W(0), W(1)` are potential
	outcomes of a binary exposure/treatment.

	Parameters
	----------
	f : function
		Function which defines the partially identified estimand.
		Must be a function of three arguments: w0, w1, y0, y1, x 
		(in that order). E.g.,
		``f = lambda w0, w1, y0, y1, x : (y0 <= y1) * (w0 <= w1)``
	outcome : np.array | pd.Series
		n-length array of outcome measurements (Y).
	instrument : np.array | pd.Series
		n-length array of binary instrument (Z).
	exposure : np.array | pd.Series
		n-length array of binary exposure (W).
	covariates : np.array | pd.Series
		(n, p)-shaped array of covariates (X).
	propensities : np.array | pd.Series
		n-length array of propensity scores :math:`P(Z=1 | X)`. 
		If ``None``, will be estimated from the data.
	clusters : np.array | pd.Series
		Optional n-length array of clusters, so ``clusters[i] = j``
		indicates that observation i is in cluster j.
	outcome_model : str | dist_reg.DistReg | list
		The model for estimating the law of :math:`Y | X, W, Z`.
		Three options:

		- A str identifier, e.g., 'ridge', 'lasso', 'elasticnet', 'randomforest', 'knn'.
		- An object inheriting from ``dist_reg.DistReg``. 
		- A list of ``dist_reg.DistReg`` objects to automatically choose between.

		E.g., when ``outcome`` is continuous, the default is
		``outcome_model=dist_reg.CtsDistReg(model_type='ridge')``.
	exposure_model : str | dist_reg.DistReg
		The model for estimating the law of :math:`W | X, Z`.
		Two options:

		- A str identifier, e.g., 'ridge', 'lasso', 'elasticnet', 'randomforest', 'knn'.
		- An object inheriting from ``dist_reg.DistReg``.
		The default is
		``exposure_model=dist_reg.BinaryDistReg(model_type='ridge')``.

	propensity_model : str | sklearn classifier
		How to estimate the propensity scores if they are not provided.
		Two options:

		- A str identifier, e.g., 'ridge', 'lasso', 'elasticnet', 'randomforest', 'knn'.
		- An sklearn classifier, e.g., ``sklearn.linear_model.LogisticRegressionCV()``.
	model_selector : dist_reg.ModelSelector
		A ModelSelector object which can choose between several outcome models.
		The default performs within-fold nested cross-validation. Note: this
		argument is ignored unless ``outcome_model`` is a list.
	discrete : bool
		If True, treats the outcome as a discrete variable. 
		Defaults to ``None`` (inferred from the data).
	support : np.array
		Optional support of the outcome, if known and discrete.
		Defaults to ``None`` (inferred from the data).
	support_restriction : function
		Boolean-valued function of w0, w1, y0, y1, x where 
		``support_restriction(w0, w1, y0, y1, x) = False`` asserts that 
		w0, w1, y0, y1, x is not in the support of 
		:math:`W(0), W(1), Y(0), Y(1), X`.
		Defaults to ``None`` (no a-priori support restrictions).
		See the user guide for important usage tips.
	model_kwargs : dict
		Additional kwargs for the ``outcome_model``, e.g.,
		``feature_transform``. See 
		:class:`dualbounds.dist_reg.CtsDistReg` or 
		:class:`dualbounds.dist_reg.BinaryDistReg` for more kwargs.
	suppress_iv_warning : bool
		If True, suppresses the beta warning for DualIVBounds.

	Notes
	-----
	This is currently slower than the ``DualBounds`` class.
	"""

	def __init__(
		self,
		exposure: Union[np.array, pd.Series],
		instrument: Union[np.array, pd.Series],
		exposure_model: Union[str, dist_reg.DistReg]='ridge',
		suppress_iv_warning: bool=False,
		*args, 
		**kwargs,
	):
		if not suppress_iv_warning:
			print("Note: DualIVBounds are in beta. Feedback is welcome. To suppress this warning, set suppress_iv_warning=True.")
		# Initialize base class
		super().__init__(
			treatment=exposure,
			*args,
			**kwargs,
		)
		# Save instrument
		if isinstance(instrument, pd.Series) or isinstance(instrument, pd.DataFrame):
			instrument = instrument.values.reshape(self.n)
		self.Z = utilities._binarize_variable(instrument, var_name='instrument')
		self.exposure_model = exposure_model

	def _solve_single_instance_cp(
		self,
		probs0: np.array,
		probs1: np.array,
		fvals: np.array,
		not_binding: np.array,
		lower: np.array,
		objval_ot: float,
		solver='CLARABEL',
	):
		"""
		Fallback solver using CP to minimize a norm-penalized variant.

		Parameters
		----------
		probs0 : array
			nvals0-length array of marginal probabilities
		probs1 : array
			nvals1-length array of marginal probabilities
		fvals : array
			(nvals0, nvals1)-length array of constraint values
		lower : bool
			If True, solves the problem for the lower CI. Else
			solves the problem for the upper CI.
		tol : float
			Ensure we are within tol of the objective.
		not_binding : array
			(nvals0, nvals1)-length boolean array. True indicates
			that something is not binding, typically because it 
			violates an a-priori constraint or is a "padded" value.

		Returns
		-------
		objval : float
			Objective value
		nu0 : array
			nvals0-length array of dual variables
		nu1 : array
			nvals1-length array of dual variables

		Notes
		-----
		This should be adapted/moved to dualbounds.generic. 
		"""
		nvals0, nvals1 = len(probs0), len(probs1)
		nu0 = cp.Variable(nvals0)
		nu1 = cp.Variable(nvals1)
		objval_unreg = cp.Parameter()
		objval_unreg.value = objval_ot
		# Constraints
		nusum = cp.reshape(nu0, (-1, 1)) + cp.reshape(nu1, (1, -1))
		flags = (~not_binding).astype(int)
		if lower:
			constraints = [
				cp.multiply(nusum, flags) <= cp.multiply(fvals, flags)
			]
		else:
			constraints = [
				cp.multiply(nusum, flags) >= cp.multiply(fvals, flags)
			]

		# Stage 1: maximize linobj subject to constraints
		linobj = nu0 @ probs0 + nu1 @ probs1
		# if lower:
		# 	lp = cp.Problem(cp.Maximize(linobj), constraints=constraints)
		# else:
		# 	lp = cp.Problem(cp.Minimize(linobj), constraints=constraints)
		# lp.solve(solver=solver)
		# print(lp.value, lp.status)
		# objval_unreg.value = lp.value

		# Stage 2: find min norm optimal solution
		norm2 = cp.norm2(nu0) + cp.norm2(nu1)
		# Constraints on true objective value
		if lower:
			constraints.append(linobj >= objval_unreg)
		else:
			constraints.append(linobj <= objval_unreg)
		# Minimize norm subject to constraints
		problem = cp.Problem(cp.Minimize(norm2), constraints=constraints)
		# Back off until we find a feasible problem (this is for numerical stability)
		for tol in [0, 1e-7, 1e-5, 1e-4, 1e-3, 1e-1, 0, 1, 3]:
			if lower:
				objval_unreg.value -= tol
			else:
				objval_unreg.value += tol
			try:
				problem.solve(solver=solver)
			except cp.SolverError as e:
				problem.solve(solver='ECOS')

			if problem.status in ['optimal', 'optimal_inaccurate']:
				break
		# Return values
		return nu0.value @ probs0 + nu1.value @ probs1, nu0.value, nu1.value

	def _solve_single_instance(
		self,
		yvals: np.array,
		probs: np.array,
		x: np.array,
		i: Optional[int]=None,
		**kwargs,
	):
		"""
		Generic instance solver.

		Parameters
		----------
		probs : np.array
			(2, 2, nvals)-shaped array where probs[z, w, y] =
			P(W(z) = w, Y(w) = yvals[z, w, y] | X = x).
		yvals : np.array
			(2, 2, nvals)-shaped array where
			yvals[z, w] is the support of Y(w) | W(z) = w
		x : np.array
			p-length covariate.
		i : int
			Optional index specifying which observation this 
			problem corresponds to. Used only to save the 
			adjusted variant of probs if probs is primal infeasible.

		Returns
		-------
		objvals : np.array
			array of the form np.array([lower_objval, upper_objval])
		nus : np.array
			(2, 2, 2, nvals)-shaped array where
			nus[1-lower, 0, w0] are the values of hatnu0(w0, y0)
			nus[1-lower, 1, w1] are the values of hatnu1(w1, y1)
		cs : np.array
			(2, 2)-shaped array of estimated cond means.
			The first axis corresponds to [lower, upper].
			The second axis corresponds to [z=0, z=1].
		"""
		# marginal probs of  W(k), Y(W(k)) | X
		## Reduce this problem to a regular OT problem.
		# Concatenate dists when Z = 0
		probs0 = np.concatenate([probs[0, 0], probs[0, 1]], axis=0)
		probs1 = np.concatenate([probs[1, 0], probs[1, 1]], axis=0)
		vals0 = np.concatenate([yvals[0, 0], yvals[0, 1]], axis=0)
		vals1 = np.concatenate([yvals[1, 0], yvals[1, 1]], axis=0)

		# Evaluate f(w0, w1, y0, y1, x).
		# This is not the final constraint matrix
		# and needs to be adjusted later, depending on the value of lower.
		nvals = yvals.shape[-1]
		fvals_raw = np.zeros((2*nvals, 2*nvals))
		not_in_support = np.zeros((2*nvals, 2*nvals)).astype(bool)
		for w0 in [0,1]:
			w0_vals = np.arange(nvals) + w0 * nvals
			for w1 in [0,1]:
				w1_vals = np.arange(nvals) + w1 * nvals
				block_inds = np.ix_(w0_vals, w1_vals)
				block_args=dict(
					w0=w0,
					w1=w1,
					y0=yvals[0, w0],
					y1=yvals[1, w1],
					x=x,
				)
				fvals_raw[block_inds] = self._apply_ot_fn(fn=self.f, **block_args)
				if self.support_restriction is not None:
					not_in_support[block_inds] = ~self._apply_ot_fn(
						fn=self.support_restriction, **block_args
					).astype(bool)

		# Signal that the off-diagonal entries of the
		# diagonal block matrices are not binding 
		# (these are essentially just padding)
		not_binding = not_in_support.copy()
		for w in [0,1]:
			inds = np.arange(nvals) + w * nvals
			block_inds = np.ix_(inds, inds)
			not_binding[block_inds] = ((not_binding[block_inds]) | ~(np.eye(nvals).astype(int))) 

		# Compute dual variables
		nu0s = np.zeros((2, 2, nvals))
		nu1s = np.zeros((2, 2, nvals))
		c0s = np.zeros(2)
		c1s = np.zeros(2)
		fvals_adj = dict()
		objvals = dict()
		for lower in [1, 0]:
			# Get rid of constraints which conflict with a-priori
			# support restrictions
			fvals = fvals_raw.copy()
			almost_inf = (2*lower - 1) * (np.abs(fvals).max() + 1) * 1e4
			fvals[not_in_support] = almost_inf
			# Adjust constraint matrix since when w0==w1,
			# we have a linear non-OT constraint. 
			# We use a hack to encode this within the OT framework
			# so that we can use the efficient OT functions.
			for w in [0,1]:
				winds = np.arange(nvals) + w * nvals
				block_inds = np.ix_(winds, winds)
				fblock = fvals[block_inds]
				if lower:
					fblock = fblock.min(axis=1-w)
				else:
					fblock = fblock.max(axis=1-w)
				# A hack to make the non-diagonal constraints unimportant
				fblock = np.ones((nvals, nvals)) * almost_inf + np.diag(fblock - almost_inf)
				fvals[block_inds] = fblock

			# Solve
			if not lower:
				fvals = -1 * fvals
			fvals_adj[lower] = fvals

		# First attempt. These dicts map lower --> output
		objvals = dict()
		logs = dict()
		for lower in [1, 0]:
			objvals[lower], logs[lower] = ot.lp.emd2(
				a=probs0,
				b=probs1,
				M=fvals_adj[lower],
				log=True,
			)

		# Check if we the "almost_inf" constraints are binding.
		# If so, the primal problem is likely infeasible and
		# the dual problem is likely unbounded. Then we resolve 
		# both the lower and upper problem.
		max_dualval = _compute_max_dualval(logs)
		if max_dualval > 0.1 * np.abs(almost_inf):
			# Ensure primal feasibility and save results
			probs = self._ensure_primal_feasibility(
				jointprobs=probs, yvals=yvals, x=x,
			)
			probs0 = np.concatenate([probs[0, 0], probs[0, 1]], axis=0)
			probs1 = np.concatenate([probs[1, 0], probs[1, 1]], axis=0)
			if i is not None:
				self._adj_jointprobs[i] = probs

			# Resolve
			for lower in [1, 0]:
				objvals[lower], logs[lower] = ot.lp.emd2(
					a=probs0,
					b=probs1,
					M=fvals_adj[lower],
					log=True,
				)

		# Multiply dual vars and objective val by -1 for upper CI
		objvals[False] *= -1
		logs[False]['u'] *= -1
		logs[False]['v'] *= -1
		# Reset fvals for  cvxpy
		fvals_adj[False] *= -1

		# If the dual variables are still too large, solve a norm penalized
		# variant with cvxpy. This is quite slow but ideally will only
		# need to be done for a few observations.
		max_dualval = _compute_max_dualval(logs)
		if max_dualval > 0.1 * np.abs(almost_inf):
			for lower in [1, 0]:
				# Resolve with CP and l2 norm
				objvals[lower], nu0, nu1 = self._solve_single_instance_cp(
					probs0=probs0,
					probs1=probs1,
					fvals=fvals_adj[lower],
					not_binding=not_binding,
					objval_ot=objvals[lower],
					lower=lower,
				)
				logs[lower]['u'] = nu0
				logs[lower]['v'] = nu1

		# Extract output
		objvals_array = np.zeros(2)
		for lower in [0,1]:
			# Extract objective value
			objvals_array[int(1-lower)] = objvals[lower]
			# Extract dual variables
			nu0s[int(1-lower)] = np.stack([logs[lower]['u'][0:nvals], logs[lower]['u'][nvals:]], axis=0)
			nu1s[int(1-lower)] = np.stack([logs[lower]['v'][0:nvals], logs[lower]['v'][nvals:]], axis=0)
			# Center
			shift = nu0s[int(1-lower)].mean()
			nu1s[int(1-lower)] += shift
			nu0s[int(1-lower)] -= shift
			# Estimated conditional means
			c0s[int(1-lower)] = np.sum(nu0s[int(1-lower)] * probs[0])
			c1s[int(1-lower)] = np.sum(nu1s[int(1-lower)] * probs[1])

		# Concatenate and return
		nus = np.stack([nu0s, nu1s], axis=1)
		cs = np.stack([c0s, c1s], axis=1)

		return objvals_array, nus, cs

	def _ensure_primal_feasibility(
		self,
		yvals,
		jointprobs,
		x,
		solver='CLARABEL',
	):
		"""
		Parameters
		----------
		jointprobs : np.array
			(2, 2, nvals)-shaped array where probs[z, w, y] =
			P(W(z) = w, Y(w) = yvals[z, w, y] | X = x).
		yvals : np.array
			(2, 2, nvals)-shaped array where
			yvals[z, w] is the support of Y(w) | W(z) = w
		x : np.array
			value of x

		Returns
		-------
		probs_adj : np.array
			(2, 2, nvals)-shaped array which is as close to
			probs as possible such that primal feasibility 
			is satisfied.

		Notes
		-----
		In this context, "primal feasibility" means that 
		there exists some law of W(0), W(1), Y(0), Y(1) | X
		which induces the implied marginals of 
		W(0), Y(W(0)) | X and W(1), Y(W(1)) | X as stored
		by probs and yvals.

		Todo
		----
		This will be a bottleneck, specifically the LP. We should
		check if there are ways to speed it up / dual formulations
		(e.g. ensure dual boundedness instead of primal feasibility.)
		"""
		jointprobs = _clip_jointprobs(jointprobs)
		## Step 1: Test that Y(0) has a common support regardless of the value of Z
		for w in [0,1]:
			np.testing.assert_array_almost_equal(
				yvals[0][w], 
				yvals[1][w],
				decimal=8,
				err_msg=f"yvals[0][{w}] != yvals[1][{w}], making the primal infeasible and the dual unbounded."
			)

		# These should be the same by default but this allows custom user inputs.
		nvals0 = len(yvals[0][0])
		nvals1 = len(yvals[1][1])

		## Step 2: set up the LP
		## Step 2(a): define core variables
		# w0yw0_probs[w][i] is the P(W(0) = w, Y(w) = yvals[0/1][w][i])
		w0yw0_probs = [cp.Variable(nvals0, pos=True), cp.Variable(nvals1, pos=True)]
		# w1yw1_probs[w][i] is the P(W(1) = w, Y(w) = yvals[0/1][w][i])
		w1yw1_probs = [cp.Variable(nvals0, pos=True), cp.Variable(nvals1, pos=True)]
		# jp_cp[w0][w1][i, j] is the joint PMF of W(0), W(1), Y(0) (value i), Y(1) (value j)
		jp_cp = [[], []]
		for w0 in [0,1]:
			for w1 in [0,1]:
				jp_cp[w0].append(cp.Variable((nvals0, nvals1), pos=True))
		# All distributions sum to 1
		constraints = [
			cp.sum(w0yw0_probs[0]) + cp.sum(w0yw0_probs[1]) == 1,
			cp.sum(w1yw1_probs[0]) + cp.sum(w1yw1_probs[1]) == 1,
			cp.sum(jp_cp[0][0] + jp_cp[0][1] + jp_cp[1][0] + jp_cp[1][1]) == 1,
		]
		## Step 2(b): support restrictions on the joint law
		if self.support_restriction is not None:
			for w0 in [0,1]:
				for w1 in [0,1]:
					block_args = dict(
						w0=w0, w1=w1, y0=yvals[0][w0], y1=yvals[1][w1], x=x,
					)
					# Account for support restrictions
					not_in_support = ~(self._apply_ot_fn(
						fn=self.support_restriction, **block_args
					).astype(bool))
					if np.any(not_in_support):
						constraints.append(jp_cp[w0][w1][not_in_support] == 0)

		### Step 2(c): marginal laws compatible with joint law
		# Note: this is not a typical OT problem! The axis choice
		# is subtle and important.
		constraints.extend([
			### 2(c)(i): the law of Y(W(0)), W(0) matches
			# W(0)=0 implies Y(W(0)) = Y(0), sum over W(1) and Y(1)
			cp.sum(jp_cp[0][0] + jp_cp[0][1], axis=1) == w0yw0_probs[0],
			# W(0) = 1 implies Y(W(0)) = Y(1), sum over W(1) and Y(0)
			cp.sum(jp_cp[1][0] + jp_cp[1][1], axis=0) == w0yw0_probs[1],
			### 2(c)(ii): the law of Y(W(1)), W(1) matches; repeat
			# W(1) = 0 implies Y(W(1)) = Y(0), sum over W(0) and Y(1)
			cp.sum(jp_cp[0][0] + jp_cp[1][0], axis=1) == w1yw1_probs[0],
			# W(1) = 1 implies Y(W(1)) = Y(1), sum over W(0) and Y(0)
			cp.sum(jp_cp[0][1] + jp_cp[1][1], axis=0) == w1yw1_probs[1],
		])

		### 3. Objective: 1-wasserstein distance over 
		# conditional laws of Y(w) | W(z) = w
		obj = 0
		for z, mprobs in zip([0,1], [w0yw0_probs, w1yw1_probs]):
			for w in [0,1]:
				### 3(a). Ensure P(W(z)  = w | Z = z) is preserved
				pw_given_z = jointprobs[z, w].sum()
				constraints.append(cp.sum(mprobs[w]) == pw_given_z)
				### 3(b). Compute Wasserstein distances
				cprobs_cp = cp.cumsum(mprobs[w])[:-1]
				cprobs_orig = np.cumsum(jointprobs[z][w])[:-1]
				ydiffs = yvals[z][w][1:] - yvals[z][w][:-1]
				obj += cp.sum(cp.multiply(ydiffs, cp.abs(cprobs_orig - cprobs_cp)))

		### 4. Create problem and solve
		problem = cp.Problem(cp.Minimize(obj), constraints)
		try:
			problem.solve(solver=solver)
		except cp.SolverError as e:
			problem.solve(solver='ECOS', verbose=True)

		### 5. Concatenate
		new_jointprobs = np.stack(
			[
				# np.stack([w0yw0_probs[0].value, w0yw0_probs[1].value], axis=0),
				# np.stack([w1yw1_probs[0].value, w1yw1_probs[1].value], axis=0),
				np.stack(
					[
						np.sum(jp_cp[0][0].value + jp_cp[0][1].value, axis=1),
						np.sum(jp_cp[1][0].value + jp_cp[1][1].value, axis=0)
					],
					axis=0
				),
				np.stack(
					[
						np.sum(jp_cp[0][0].value + jp_cp[1][0].value, axis=1),
						np.sum(jp_cp[0][1].value + jp_cp[1][1].value, axis=0)
					],
					axis=0
				),
			],
			axis=0
		)
		### 6. Clip and ensure sums to one
		# new_jointprobs = np.clip(new_jointprobs, 0, 1)
		# for z in [0,1]:
		# 	new_jointprobs[z] /= new_jointprobs[z].sum()
		return new_jointprobs

	def _ensure_feasibility(
		self,
		i,
		orig_nus,
		lower,
		ymin,
		ymax,
		grid_size=100,
		tol=5e-4,
	):
		"""
		Parameters
		----------
		i : int
			Index of which data-point we are performing
			this operation for.
		orig_nus : np.array
			(2, 2, nvals)-length array of dual variables 
			First axis is associated with [z=0, z=1]
			Second axis is associated with [w=0, w=1]
			Third axis is associated with [self._yvals[i, z, w]]
		lower : bool
			Specifies lower vs. upper bound.
		ymin : float
			Minimum value of y
		ymax : float
			Maximum value of y
		grid_size : int
			Grid size along each dimension.

		Returns
		-------
		new_yvals : list
			new_yvals[z][w] contains a  ``nvals + grid_size`` length
			array of new yvals.
			*Exact size may change to avoid duplicate values
		new_nus : list
			new_nus are the dual variables for new_yvals.
		dx : float
			Maximum numerical error induced by interpolation
			process.
		objval_diff : np.array
			The estimated objective value change for the new
			dual variables.
		"""
		# original objective value
		probs = self._jointprobs[i]
		obj_orig = np.sum(probs * orig_nus)
		if grid_size > 0:
			# Construct new yvals
			extra_yvals = np.linspace(ymin, ymax, grid_size)
			new_yvals = [[], []]
			for z in [0,1]:
				for w in [0,1]:
					# yvals
					new_yvals[z].append(np.sort(np.unique(np.concatenate(
						[self._yvals[i, z, w], extra_yvals], axis=0
					))))
			# New fvals
			new_fvals = [[], []]
			for w0 in [0,1]:
				for w1 in [0,1]:
					block_args = dict(
						w0=w0,
						w1=w1,
						y0=new_yvals[0][w0],
						y1=new_yvals[1][w1],
						x=self.X[i],
					)
					# fvals
					new_fvals[w0].append(self._apply_ot_fn(fn=self.f, **block_args))
					# Account for support restrictions
					if self.support_restriction is not None:
						not_in_support = ~(self._apply_ot_fn(
							fn=self.support_restriction, **block_args
						).astype(bool))
					else:
						not_in_support = np.zeros(new_fvals[w0][-1].shape).astype(bool)
					if lower:
						new_fvals[w0][-1][not_in_support] = np.inf
					else:
						new_fvals[w0][-1][not_in_support] = -np.inf

			# Interpolate nus
			new_nus = [[], []]
			for z in [0,1]:
				for w in [0,1]:
					new_nus[z].append(self.interp_fn(
						x=self._yvals[i, z, w], y=orig_nus[z, w], newx=new_yvals[z][w]
					))
			### Check feasibility. Two cases.
			### Case 1: w0 == w1. Here we have vanilla linear (non-OT) constraints
			for w in [0,1]:
				if lower:
					new_fvals_block = new_fvals[w][w].min(axis=1-w)
				else:
					new_fvals_block = new_fvals[w][w].max(axis=1-w)
				deltas = new_nus[0][w] + new_nus[1][w] - new_fvals_block
				if lower:
					deltas = np.maximum(deltas, 0)
					new_nus[0][w] -= deltas / 2
					new_nus[1][w] -= deltas / 2
				else:
					deltas = np.minimum(deltas, 0)
					new_nus[0][w] -= deltas / 2
					new_nus[1][w] -= deltas / 2

			### Case 2: w0 != w1. Here we have full OT constraints.
			for w0, w1 in zip([0,1], [1,0]):
				deltas = new_nus[0][w0].reshape(-1, 1) + new_nus[1][w1].reshape(1, -1)
				deltas = deltas - new_fvals[w0][w1]
				# Figure out which axis to adjust
				if lower:
					deltas = np.maximum(deltas, 0)
					deltas0 = deltas.max(axis=1)
					dx0 = np.mean(np.maximum(deltas0, 0))
					deltas1 = deltas.max(axis=0)
					dx1 = np.mean(np.maximum(deltas1, 0))
					adj_axis = 1 if dx1 <= dx0 else 0
				else:
					deltas = np.minimum(deltas, 0)
					deltas0 = deltas.min(axis=1)
					dx0 = np.mean(np.minimum(deltas0, 0))
					deltas1 = deltas.min(axis=0)
					dx1 = np.mean(np.minimum(deltas1, 0))
					adj_axis = 1 if dx1 >= dx0 else 0
				# Adjust
				if adj_axis == 0:
					new_nus[0][w0] -= deltas0
				else:
					new_nus[1][w1] -= deltas1
			
			# Compute difference in objective values by remapping to original shape
			adj_nus = np.zeros(orig_nus.shape)
			for z in [0,1]:
				for w in [0,1]:
					adj_nus[z, w] = self.interp_fn(
						x=new_yvals[z][w], y=new_nus[z][w], newx=self._yvals[i, z, w],
					)
			objval_diff = obj_orig - np.sum(adj_nus * probs)
		else:
			new_yvals = self._yvals[i]
			new_nus = orig_nus
			objval_diff = 0
		# return
		return new_yvals, new_nus, objval_diff

	def _interpolate_and_ensure_feas(
		self, yi, wi, zi, lower, i, ymin, ymax, **kwargs
	):  
		"""
		This is used in two places:
		1. _compute_realized_dual_variables
		2. Main loop of compute_dual_variables

		We use this in the main loop of compute_dual_variables
		instead of just calling _compute_realized_dual_variables
		so that the progress report timing is accurate.
		"""
		nusx = self.nus[i, 1-lower]
		# Ensure feasibilitynus=nusx,
		if not self.discrete:
			new_yvals, new_nus, odx = self._ensure_feasibility(
				i=i, 
				orig_nus=nusx, 
				lower=lower, 
				ymin=ymin,
				ymax=ymax, 
				**kwargs
			)
			self.objdiffs[1-lower, i] = odx
			# interpolate to find realized values 
			self.hatnus[1-lower, i] = self.interp_fn(
				x=new_yvals[zi][wi], y=new_nus[zi][wi], newx=yi,
			).item()
		else:
			j = np.where(self._yvals[i, zi, wi] == yi)[0][0]
			self.hatnus[1-lower, i] = nusx[zi, wi, j]

	def compute_dual_variables(
		self,
		wprobs: np.array,
		ydists: list,
		ymin: Optional[float]=None,
		ymax: Optional[float]=None,
		min_quantile: Optional[float]=None,
		interp_fn: callable=interpolation.adaptive_interpolate,
		nvals: int=100,
		ninterp: Optional[int]=None,
		verbose: bool=True,
		**kwargs
	):
		"""
		Same signature as `generic.DualBounds.compute_dual_variables`
		with the following exceptions.

		Parameters
		----------
		wprobs : np.array
			wprobs[i, z] is the estimated probability 
			:math:`P(W(z) = i | X_i)`
		ydists : list
			For z, w in {0,1}, ydists[z][w] must be a list of batched
			scipy dists whose shapes sum to n. The ith element 
			of ydists[z][w] represents the law of 
			:math:`Y_i(w) | W_i(z) = w, X_i`.
		min_quantile : float
			Minimum quantile to consider when discretizing. 
			Defaults to 1 / (2*nvals).
		"""
		# Setup
		if verbose:
			print("Estimating optimal dual variables.")
		self.interp_fn = interp_fn
		if self.discrete:
			self.nvals = len(self.support)
		else:
			self.nvals = nvals
			##TODO?
			# if self.nvals <= generic.MIN_NVALS

		# Infer ymin/ymax
		if ymin is None:
			ymin = self.y.min() - (self.y.max() - self.y.min()) / 2
		if ymax is None:
			ymax = self.y.max() + (self.y.max() - self.y.min()) / 2

		# Discretize if necessary
		self._yvals = [[], []]
		self._yprobs = [[], []]
		for z in [0,1]:
			for w in [0,1]:
				yvals_zw, yprobs_zw = self._discretize(
					ydists[z][w], 
					nvals=self.nvals, 
					min_quantile=min_quantile,
					ymin=ymin,
					ymax=ymax,
					ninterp=ninterp,
				)
				# ensure everything is sorted across the nvals axis
				yvals_zw, yprobs_zw = utilities._sort_disc_dist(
					vals=yvals_zw, probs=yprobs_zw
				)
				self._yvals[z].append(yvals_zw)
				self._yprobs[z].append(yprobs_zw)

		# Ensure that the support of Y(w) | W(0) = w and Y(w) | W(1) = w 
		# is the same. This is necessary to ensure primal feasibility.
		concat_ys = False
		for w in [0,1]:
			if not np.all(self._yvals[0][w] == self._yvals[1][w]):
				concat_ys = True
		# Concatenate
		if concat_ys:
			for w in [0,1]:
				# zero padding for probabilities
				zero_pad = np.zeros((self.n, self.nvals))
				# Create the union of the supports
				yvals_w = np.concatenate([self._yvals[0][w], self._yvals[1][w]], axis=1)
				self._yvals[0][w] = yvals_w
				self._yvals[1][w] = yvals_w
				# Pad probabilities
				self._yprobs[0][w] = np.concatenate([self._yprobs[0][w], zero_pad], axis=1)
				self._yprobs[1][w] = np.concatenate([zero_pad, self._yprobs[1][w]], axis=1)
				# argsort
				for z in [0,1]:
					self._yvals[z][w], self._yprobs[z][w] = utilities._sort_disc_dist(
						vals=self._yvals[z][w], probs=self._yprobs[z][w]
					)

			# Adjust self.nvals
			self.nvals *= 2

		# Shape: (n, 2, 2, nvals) with the 2s corresponding to (z,w) in that order.
		self._yvals = np.stack(
			[np.stack(self._yvals[0], axis=1), np.stack(self._yvals[1], axis=1)],
			axis=1
		)
		self._yprobs = np.stack(
			[np.stack(self._yprobs[0], axis=1), np.stack(self._yprobs[1], axis=1)],
			axis=1
		)

		# joint law of (W, Y(W)) | Z
		self._jointprobs = self._yprobs.copy()
		self._jointprobs[:, :, 0] *= (1 - wprobs.reshape(self.n, 2, 1))
		self._jointprobs[:, :, 1] *= wprobs.reshape(self.n, 2, 1)
		self._adj_jointprobs = self._jointprobs.copy()

		## Initialize results
		# first dimension = [lower, upper]
		# second dimension = [z=0, z=1]
		# third dimension = [w=0, w=1]
		self.nus = np.zeros((self.n, 2, 2, 2, self.nvals))
		# realized dual variables
		self.hatnus = np.zeros((2, self.n))
		self.objvals = np.zeros((2, self.n))
		self.objdiffs = np.zeros((2, self.n))
		# estimated conditional means of nus given Z (NOT given W)
		# first dimension = [lower, upper]
		# second dimension = [z=0, z=1]
		self.cs = np.zeros((self.n, 2, 2)) 
		for i in utilities.vrange(self.n, verbose=verbose):
			# Compute optimal dual variable functions
			objvalsx, nusx, csx = self._solve_single_instance(
				yvals=self._yvals[i],
				probs=self._jointprobs[i],
				x=self.X[i],
				ymin=ymin,
				ymax=ymax,
				i=i,
			)
			self.objvals[:, i] = objvalsx
			self.nus[i] = nusx
			self.cs[i] = csx
			for lower in [0,1]:
				# Compute realized dual variables
				self._interpolate_and_ensure_feas(
					yi=self.y[i],
					zi=self.Z[i],
					wi=self.W[i],
					lower=lower,
					ymin=ymin,
					ymax=ymax,
					i=i,
					**kwargs,
				)

	def _compute_realized_dual_variables(self, y=None, Z=None, W=None, **kwargs):
		"""
		Helper function which applies interpolation 
		(for continuous Y) to compute realized dual variables.
		It also ensures feasibility through a 2D gridsearch.
		This is used primarily in unit tests.
		"""
		y = self.y if y is None else y
		W = self.W if W is None else W
		Z = self.Z if Z is None else Z

		# Create ylims
		ymin = kwargs.pop("ymin", y.min() - (y.max() - y.min()) / 2)
		ymax = kwargs.pop("ymax", y.max() + (y.max() - y.min()) / 2)

		### Compute realized hatnu1s/hatnu0s
		for i in range(self.n):
			for lower in [0, 1]:
				self._interpolate_and_ensure_feas(
					i=i,
					yi=y[i],
					zi=Z[i],
					wi=W[i],
					lower=lower, 
					ymin=ymin,
					ymax=ymax,
					**kwargs
				)

	def _compute_ipw_summands(self):
		"""
		Helper method to compute (A)IPW estimator summands.
		Must be run after running ``self.compute_dual_variables``
		"""
		# initialize outputs
		self.ipw_summands = np.zeros((2, self.n))
		self.aipw_summands = np.zeros((2, self.n))

		# Compute IPW summands and AIPW summands
		Z = self.Z.reshape(1, self.n)
		pis = self.pis.reshape(1, self.n)
		ipw_denom = Z * (pis) + (1 - Z) * (1 - pis) 
		self.ipw_summands = self.hatnus / ipw_denom
		# Compute AIPW summands
		mus = self.cs.sum(axis=2).T # (2, n)-shaped array
		deltas = self.cs[:, :, 1].T * Z + self.cs[:, :, 0].T * (1 - Z) # (2, n)-shaped array
		self.aipw_summands = (self.hatnus - deltas) / ipw_denom + mus

	def _compute_cond_means(self):
		"""
		Computes self.mu0 = E[Y(W(0)) | Z=0]
		and self.mu1 = E[Y(W(1)) | Z = 1].
		"""
		ydist_means = np.zeros((2, 2, self.n))
		for z in [0,1]:
			for w in [0,1]:
				dists_zw = self.ydists[z][w]
				if not isinstance(dists_zw, list):
					dists_zw = [dists_zw]
				ydist_means[z, w] = np.concatenate([x.mean() for x in dists_zw])

		self.mu0 = self.wprobs[:, 0] * ydist_means[0, 1]
		self.mu0 += (1 - self.wprobs[:, 0]) * ydist_means[0][0]
		self.mu1 = self.wprobs[:, 1] * ydist_means[1][1]
		self.mu1 += (1 - self.wprobs[:, 1]) * ydist_means[1][0]

	def _compute_oos_resids(self):
		"""
		Computes out-of-sample predictions of Y | Z and residuals.
		"""
		self._compute_cond_means()
		self.oos_preds = self.mu0.copy()
		self.oos_preds[self.Z == 1] = self.mu1[self.Z == 1]
		# residuals and return
		self.oos_resids = self.y - self.oos_preds
		return self.oos_resids

	def cross_fit(
		self,
		nfolds: int=5,
		suppress_warning: bool=False,
		verbose: bool=True,
	):
		"""
		Cross-fits the outcome model.

		Parameters
		----------
		nfolds : int
			Number of folds to use in cross-fitting.
		suppress_warning : bool
			If True, suppresses a potential warning about cross-fitting.
		verbose : bool
			If True, prints progress reports.

		Returns
		-------
		ydists : list
			for z, w in [0,1], ydists[z][w] is a list of batched scipy
			distributions whose shapes sum to n. the ith distribution 
			is the out-of-sample estimate of the conditional law
			of :math:`Y_i(w) | W(z) = w, X_i`.
		"""
		# if pis not supplied: will use cross-fitting
		if self.pis is None:
			self.fit_propensity_scores(nfolds=nfolds, verbose=verbose)

		# Fit exposure model
		if self.wprobs is None:
			self.exposure_model = generic.get_default_model(
				outcome_model=self.exposure_model, 
				# the following args are ignored if exposure_model
				# inherits from dist_reg.DistReg class
				support=set([0,1]), 
				discrete=True,
				monotonicity=True, 
				how_transform='intercept',
			)
			if verbose:
				print("Cross-fitting the exposure model.")
			wout = dist_reg.cross_fit_predictions(
				W=self.Z, X=self.X, y=self.W, 
				nfolds=nfolds, 
				model=self.exposure_model,
				verbose=verbose,
				probs_only=True,
			)
			wprobs, self.exposure_model_fits, _ = wout
			self.wprobs = np.stack(wprobs, axis=1)
		elif not suppress_warning:
			warnings.warn(generic.CROSSFIT_WARNING.replace("y0_dists/y1_dists", "wprobs"))

		# Fit outcome model
		if self.ydists is None:
			# Note: this returns the existing model
			# if an existing model is provided
			self.outcome_model = generic.get_default_model(
				discrete=self.discrete, 
				support=self.support,
				outcome_model=self.outcome_model,
				**self.model_kwargs
			)
			if verbose:
				print("Cross-fitting the outcome model.")
			y_out = dist_reg.cross_fit_predictions(
				W=self.W, X=self.X, y=self.y, Z=self.Z, 
				nfolds=nfolds, 
				model=self.outcome_model,
				verbose=verbose,
			)
			self.ydists, self.model_fits, self.oos_dist_preds = y_out
		elif not suppress_warning:
			warnings.warn(generic.CROSSFIT_WARNING)

		return self.ydists


	def fit(
		self,
		nfolds: int = 5,
		aipw: bool = True,
		alpha: float = 0.05,
		wprobs: Optional[list]=None,
		ydists: Optional[list]=None,
		verbose: bool = True,
		suppress_warning: bool = False,
		**solve_kwargs,
	):
		# Fit model of W | X and Y | X if not provided
		self.ydists, self.wprobs = ydists, wprobs
		self.cross_fit(
			nfolds=nfolds, suppress_warning=suppress_warning, verbose=verbose,
		)

		# compute dual variables
		self.compute_dual_variables(
			ydists=self.ydists,
			wprobs=self.wprobs,
			verbose=verbose,
			**solve_kwargs,
		)
		# compute dual bounds
		self.alpha = alpha
		self._compute_final_bounds(aipw=aipw, alpha=alpha)
		return self

	def eval_treatment_model():
		return dist_reg._evaluate_model_predictions(
			y=self.Z, haty=self.pis
		)

	def eval_exposure_model():
		"""
		Thinly wraps dist_reg._evaluate_model_predictions.

		Returns
		-------
		sumstats : pd.DataFrame
			DataFrame summarizing goodness-of-fit metrics for
			the cross-fit exposure scores.
		"""
		return dist_reg._evaluate_model_predictions(
			y=self.W, 
			haty=self.wprobs[(np.arange(self.n), self.Z.astype(int))]
		)


	def summary(self, minval=-np.inf, maxval=np.inf):
		print("___________________Inference_____________________")
		print(self.results(minval=minval, maxval=maxval))
		print()
		print("_________________Outcome model___________________")
		print(self.eval_outcome_model())
		print()
		print("_________________Exposure model__________________")
		print(self.eval_exposure_model())
		print()
		print("_________________Treatment model_________________")
		print(self.eval_treatment_model())
		print(sumstats)
		print()